"""
.. include:: ../README.md

# API Documentation
"""

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = "0.26.0"

import sys

if sys.version_info < (3, 10):  # pragma: no cover
    raise RuntimeError("Requires Python 3.10 or later.")

from .controller import Controller
from .gnmiclient import GNMIClient, GNMISubscription, GNMIUpdate
from .gnmipath import GNMIPath
from .grpcutil import GRPCCredentialsTLS, GRPCStatusCode
from .log import LoggerAdapter
from .macaddr import MACAddress
from .p4client import P4Client, P4ClientError, P4Error
from .p4entity import (
    P4ActionProfileGroup,
    P4ActionProfileMember,
    P4CloneSessionEntry,
    P4CounterData,
    P4CounterEntry,
    P4DigestEntry,
    P4DigestList,
    P4DigestListAck,
    P4DirectCounterEntry,
    P4DirectMeterEntry,
    P4ExternEntry,
    P4IndirectAction,
    P4Member,
    P4MeterConfig,
    P4MeterCounterData,
    P4MeterEntry,
    P4MulticastGroupEntry,
    P4PacketIn,
    P4PacketOut,
    P4RegisterEntry,
    P4TableAction,
    P4TableEntry,
    P4TableMatch,
    P4ValueSetEntry,
)
from .p4schema import P4ConfigAction, P4CounterUnit, P4Schema
from .ports import SwitchPort, SwitchPortList
from .runner import run
from .switch import Switch, SwitchEvent, SwitchOptions

Match = P4TableMatch
"`Match` is an alias for P4TableMatch."

Action = P4TableAction
"`Action` is an alias for P4TableAction."

IndirectAction = P4IndirectAction
"`IndirectAction` is an alias for P4IndirectAction."

__all__ = [
    "run",
    "Controller",
    "LoggerAdapter",
    "MACAddress",
    "P4ActionProfileGroup",
    "P4ActionProfileMember",
    "P4Client",
    "P4ClientError",
    "P4CloneSessionEntry",
    "P4CounterData",
    "P4CounterEntry",
    "P4CounterUnit",
    "P4DigestEntry",
    "P4DigestList",
    "P4DigestListAck",
    "P4DirectCounterEntry",
    "P4DirectMeterEntry",
    "P4Error",
    "P4ExternEntry",
    "IndirectAction",  # alias for P4IndirectAction
    "P4IndirectAction",
    "P4Member",
    "P4MeterConfig",
    "P4MeterCounterData",
    "P4MeterEntry",
    "P4MulticastGroupEntry",
    "P4PacketIn",
    "P4PacketOut",
    "P4RegisterEntry",
    "Action",  # alias for P4TableAction
    "P4TableAction",
    "P4TableEntry",
    "Match",  # alias for P4TableMatch
    "P4TableMatch",
    "P4ValueSetEntry",
    "P4ConfigAction",
    "P4Schema",
    "Switch",
    "SwitchEvent",
    "SwitchOptions",
    "SwitchPort",
    "SwitchPortList",
    "GNMIClient",
    "GNMIPath",
    "GNMISubscription",
    "GNMIUpdate",
    "GRPCCredentialsTLS",
    "GRPCStatusCode",
]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2022-present Open Networking Foundation
Copyright 2022 Intel Corporation
SPDX-License-Identifier: Apache-2.0
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class P4RoleConfig(google.protobuf.message.Message):
    """The P4RoleConfig message conists of these fields:
     exclusive_p4_ids - A list of P4 entities for which this role exclusivly may
         issue Write updates. Those entities are also filtered out of Read
         responses.
     shared_p4_ids - A list of P4 entities for which this role may issue Write
         updates. Access to those entities is shared with other roles. Those
         entities are also filtered out of Read responses.
     packet_in_filter - A single PacketFilter that is applied to incoming packets
         to determine whether this role should receive them. An empty list
         results in no filtering. This is a positive filter, a packet must
         contain the exact specified value to be forwarded.
     receives_packet_ins - A toggle to set if this role should receive PacketIns.
     can_push_pipeline - Determines if this role is allowed to push a pipeline.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class PacketFilter(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        METADATA_ID_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        metadata_id: builtins.int
        """Must match an ID in the P4Info."""
        value: builtins.bytes
        """Should be given in canonical form."""
        def __init__(
            self,
            *,
            metadata_id: builtins.int = ...,
            value: builtins.bytes = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["metadata_id", b"metadata_id", "value", b"value"]) -> None: ...

    EXCLUSIVE_P4_IDS_FIELD_NUMBER: builtins.int
    SHARED_P4_IDS_FIELD_NUMBER: builtins.int
    PACKET_IN_FILTER_FIELD_NUMBER: builtins.int
    RECEIVES_PACKET_INS_FIELD_NUMBER: builtins.int
    CAN_PUSH_PIPELINE_FIELD_NUMBER: builtins.int
    receives_packet_ins: builtins.bool
    can_push_pipeline: builtins.bool
    @property
    def exclusive_p4_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def shared_p4_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def packet_in_filter(self) -> global___P4RoleConfig.PacketFilter: ...
    def __init__(
        self,
        *,
        exclusive_p4_ids: collections.abc.Iterable[builtins.int] | None = ...,
        shared_p4_ids: collections.abc.Iterable[builtins.int] | None = ...,
        packet_in_filter: global___P4RoleConfig.PacketFilter | None = ...,
        receives_packet_ins: builtins.bool = ...,
        can_push_pipeline: builtins.bool = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["packet_in_filter", b"packet_in_filter"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["can_push_pipeline", b"can_push_pipeline", "exclusive_p4_ids", b"exclusive_p4_ids", "packet_in_filter", b"packet_in_filter", "receives_packet_ins", b"receives_packet_ins", "shared_p4_ids", b"shared_p4_ids"]) -> None: ...

global___P4RoleConfig = P4RoleConfig



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
This package and its contents are a work-in-progress."""

import abc
import collections.abc
import grpc
import grpc.aio
from . import p4runtime_pb2 as _dot_p4runtime_pb2
import typing

_T = typing.TypeVar("_T")

class _MaybeAsyncIterator(collections.abc.AsyncIterator[_T], collections.abc.Iterator[_T], metaclass=abc.ABCMeta): ...

class _ServicerContext(grpc.ServicerContext, grpc.aio.ServicerContext):  # type: ignore[misc, type-arg]
    ...

class P4RuntimeStub:
    def __init__(self, channel: typing.Union[grpc.Channel, grpc.aio.Channel]) -> None: ...
    Write: grpc.UnaryUnaryMultiCallable[
        _dot_p4runtime_pb2.WriteRequest,
        _dot_p4runtime_pb2.WriteResponse,
    ]
    """Update one or more P4 entities on the target."""

    Read: grpc.UnaryStreamMultiCallable[
        _dot_p4runtime_pb2.ReadRequest,
        _dot_p4runtime_pb2.ReadResponse,
    ]
    """Read one or more P4 entities from the target."""

    SetForwardingPipelineConfig: grpc.UnaryUnaryMultiCallable[
        _dot_p4runtime_pb2.SetForwardingPipelineConfigRequest,
        _dot_p4runtime_pb2.SetForwardingPipelineConfigResponse,
    ]
    """Sets the P4 forwarding-pipeline config."""

    GetForwardingPipelineConfig: grpc.UnaryUnaryMultiCallable[
        _dot_p4runtime_pb2.GetForwardingPipelineConfigRequest,
        _dot_p4runtime_pb2.GetForwardingPipelineConfigResponse,
    ]
    """Gets the current P4 forwarding-pipeline config."""

    StreamChannel: grpc.StreamStreamMultiCallable[
        _dot_p4runtime_pb2.StreamMessageRequest,
        _dot_p4runtime_pb2.StreamMessageResponse,
    ]
    """Represents the bidirectional stream between the controller and the
    switch (initiated by the controller), and is managed for the following
    purposes:
    - connection initiation through client arbitration
    - indicating switch session liveness: the session is live when switch
      sends a positive client arbitration update to the controller, and is
      considered dead when either the stream breaks or the switch sends a
      negative update for client arbitration
    - the controller sending/receiving packets to/from the switch
    - streaming of notifications from the switch
    """

    Capabilities: grpc.UnaryUnaryMultiCallable[
        _dot_p4runtime_pb2.CapabilitiesRequest,
        _dot_p4runtime_pb2.CapabilitiesResponse,
    ]

class P4RuntimeAsyncStub:
    Write: grpc.aio.UnaryUnaryMultiCallable[
        _dot_p4runtime_pb2.WriteRequest,
        _dot_p4runtime_pb2.WriteResponse,
    ]
    """Update one or more P4 entities on the target."""

    Read: grpc.aio.UnaryStreamMultiCallable[
        _dot_p4runtime_pb2.ReadRequest,
        _dot_p4runtime_pb2.ReadResponse,
    ]
    """Read one or more P4 entities from the target."""

    SetForwardingPipelineConfig: grpc.aio.UnaryUnaryMultiCallable[
        _dot_p4runtime_pb2.SetForwardingPipelineConfigRequest,
        _dot_p4runtime_pb2.SetForwardingPipelineConfigResponse,
    ]
    """Sets the P4 forwarding-pipeline config."""

    GetForwardingPipelineConfig: grpc.aio.UnaryUnaryMultiCallable[
        _dot_p4runtime_pb2.GetForwardingPipelineConfigRequest,
        _dot_p4runtime_pb2.GetForwardingPipelineConfigResponse,
    ]
    """Gets the current P4 forwarding-pipeline config."""

    StreamChannel: grpc.aio.StreamStreamMultiCallable[
        _dot_p4runtime_pb2.StreamMessageRequest,
        _dot_p4runtime_pb2.StreamMessageResponse,
    ]
    """Represents the bidirectional stream between the controller and the
    switch (initiated by the controller), and is managed for the following
    purposes:
    - connection initiation through client arbitration
    - indicating switch session liveness: the session is live when switch
      sends a positive client arbitration update to the controller, and is
      considered dead when either the stream breaks or the switch sends a
      negative update for client arbitration
    - the controller sending/receiving packets to/from the switch
    - streaming of notifications from the switch
    """

    Capabilities: grpc.aio.UnaryUnaryMultiCallable[
        _dot_p4runtime_pb2.CapabilitiesRequest,
        _dot_p4runtime_pb2.CapabilitiesResponse,
    ]

class P4RuntimeServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def Write(
        self,
        request: _dot_p4runtime_pb2.WriteRequest,
        context: _ServicerContext,
    ) -> typing.Union[_dot_p4runtime_pb2.WriteResponse, collections.abc.Awaitable[_dot_p4runtime_pb2.WriteResponse]]:
        """Update one or more P4 entities on the target."""

    @abc.abstractmethod
    def Read(
        self,
        request: _dot_p4runtime_pb2.ReadRequest,
        context: _ServicerContext,
    ) -> typing.Union[collections.abc.Iterator[_dot_p4runtime_pb2.ReadResponse], collections.abc.AsyncIterator[_dot_p4runtime_pb2.ReadResponse]]:
        """Read one or more P4 entities from the target."""

    @abc.abstractmethod
    def SetForwardingPipelineConfig(
        self,
        request: _dot_p4runtime_pb2.SetForwardingPipelineConfigRequest,
        context: _ServicerContext,
    ) -> typing.Union[_dot_p4runtime_pb2.SetForwardingPipelineConfigResponse, collections.abc.Awaitable[_dot_p4runtime_pb2.SetForwardingPipelineConfigResponse]]:
        """Sets the P4 forwarding-pipeline config."""

    @abc.abstractmethod
    def GetForwardingPipelineConfig(
        self,
        request: _dot_p4runtime_pb2.GetForwardingPipelineConfigRequest,
        context: _ServicerContext,
    ) -> typing.Union[_dot_p4runtime_pb2.GetForwardingPipelineConfigResponse, collections.abc.Awaitable[_dot_p4runtime_pb2.GetForwardingPipelineConfigResponse]]:
        """Gets the current P4 forwarding-pipeline config."""

    @abc.abstractmethod
    def StreamChannel(
        self,
        request_iterator: _MaybeAsyncIterator[_dot_p4runtime_pb2.StreamMessageRequest],
        context: _ServicerContext,
    ) -> typing.Union[collections.abc.Iterator[_dot_p4runtime_pb2.StreamMessageResponse], collections.abc.AsyncIterator[_dot_p4runtime_pb2.StreamMessageResponse]]:
        """Represents the bidirectional stream between the controller and the
        switch (initiated by the controller), and is managed for the following
        purposes:
        - connection initiation through client arbitration
        - indicating switch session liveness: the session is live when switch
          sends a positive client arbitration update to the controller, and is
          considered dead when either the stream breaks or the switch sends a
          negative update for client arbitration
        - the controller sending/receiving packets to/from the switch
        - streaming of notifications from the switch
        """

    @abc.abstractmethod
    def Capabilities(
        self,
        request: _dot_p4runtime_pb2.CapabilitiesRequest,
        context: _ServicerContext,
    ) -> typing.Union[_dot_p4runtime_pb2.CapabilitiesResponse, collections.abc.Awaitable[_dot_p4runtime_pb2.CapabilitiesResponse]]: ...

def add_P4RuntimeServicer_to_server(servicer: P4RuntimeServicer, server: typing.Union[grpc.Server, grpc.aio.Server]) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: p4/config/v1/p4info.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from . import p4types_pb2 as p4_dot_config_dot_v1_dot_p4types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19p4/config/v1/p4info.proto\x12\x0cp4.config.v1\x1a\x19google/protobuf/any.proto\x1a\x1ap4/config/v1/p4types.proto\"\x88\x05\n\x06P4Info\x12\'\n\x08pkg_info\x18\x01 \x01(\x0b\x32\x15.p4.config.v1.PkgInfo\x12#\n\x06tables\x18\x02 \x03(\x0b\x32\x13.p4.config.v1.Table\x12%\n\x07\x61\x63tions\x18\x03 \x03(\x0b\x32\x14.p4.config.v1.Action\x12\x34\n\x0f\x61\x63tion_profiles\x18\x04 \x03(\x0b\x32\x1b.p4.config.v1.ActionProfile\x12\'\n\x08\x63ounters\x18\x05 \x03(\x0b\x32\x15.p4.config.v1.Counter\x12\x34\n\x0f\x64irect_counters\x18\x06 \x03(\x0b\x32\x1b.p4.config.v1.DirectCounter\x12#\n\x06meters\x18\x07 \x03(\x0b\x32\x13.p4.config.v1.Meter\x12\x30\n\rdirect_meters\x18\x08 \x03(\x0b\x32\x19.p4.config.v1.DirectMeter\x12J\n\x1a\x63ontroller_packet_metadata\x18\t \x03(\x0b\x32&.p4.config.v1.ControllerPacketMetadata\x12*\n\nvalue_sets\x18\n \x03(\x0b\x32\x16.p4.config.v1.ValueSet\x12)\n\tregisters\x18\x0b \x03(\x0b\x32\x16.p4.config.v1.Register\x12%\n\x07\x64igests\x18\x0c \x03(\x0b\x32\x14.p4.config.v1.Digest\x12%\n\x07\x65xterns\x18\x64 \x03(\x0b\x32\x14.p4.config.v1.Extern\x12,\n\ttype_info\x18\xc8\x01 \x01(\x0b\x32\x18.p4.config.v1.P4TypeInfo\"3\n\rDocumentation\x12\r\n\x05\x62rief\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\"\x9c\x01\n\x12PlatformProperties\x12\"\n\x1amulticast_group_table_size\x18\x01 \x01(\x05\x12,\n$multicast_group_table_total_replicas\x18\x02 \x01(\x05\x12\x34\n,multicast_group_table_max_replicas_per_entry\x18\x03 \x01(\x05\"\xe8\x02\n\x07PkgInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12(\n\x03\x64oc\x18\x03 \x01(\x0b\x32\x1b.p4.config.v1.Documentation\x12\x13\n\x0b\x61nnotations\x18\x04 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\n \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x0c\n\x04\x61rch\x18\x05 \x01(\t\x12\x14\n\x0corganization\x18\x06 \x01(\t\x12\x0f\n\x07\x63ontact\x18\x07 \x01(\t\x12\x0b\n\x03url\x18\x08 \x01(\t\x12\x42\n\x16structured_annotations\x18\t \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\x12=\n\x13platform_properties\x18\x0b \x01(\x0b\x32 .p4.config.v1.PlatformProperties\"\x87\x02\n\x05P4Ids\"\xfd\x01\n\x06Prefix\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x41\x43TION\x10\x01\x12\t\n\x05TABLE\x10\x02\x12\r\n\tVALUE_SET\x10\x03\x12\x15\n\x11\x43ONTROLLER_HEADER\x10\x04\x12\x15\n\x11PSA_EXTERNS_START\x10\x10\x12\x12\n\x0e\x41\x43TION_PROFILE\x10\x11\x12\x0b\n\x07\x43OUNTER\x10\x12\x12\x12\n\x0e\x44IRECT_COUNTER\x10\x13\x12\t\n\x05METER\x10\x14\x12\x10\n\x0c\x44IRECT_METER\x10\x15\x12\x0c\n\x08REGISTER\x10\x16\x12\n\n\x06\x44IGEST\x10\x17\x12\x18\n\x13OTHER_EXTERNS_START\x10\x80\x01\x12\x08\n\x03MAX\x10\xff\x01\"\xf2\x01\n\x08Preamble\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05\x61lias\x18\x03 \x01(\t\x12\x13\n\x0b\x61nnotations\x18\x04 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x07 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12(\n\x03\x64oc\x18\x05 \x01(\x0b\x32\x1b.p4.config.v1.Documentation\x12\x42\n\x16structured_annotations\x18\x06 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"k\n\x06\x45xtern\x12\x16\n\x0e\x65xtern_type_id\x18\x01 \x01(\r\x12\x18\n\x10\x65xtern_type_name\x18\x02 \x01(\t\x12/\n\tinstances\x18\x03 \x03(\x0b\x32\x1c.p4.config.v1.ExternInstance\"^\n\x0e\x45xternInstance\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\"\n\x04info\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\"\xdc\x03\n\nMatchField\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x61nnotations\x18\x03 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\n \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x10\n\x08\x62itwidth\x18\x04 \x01(\x05\x12\x38\n\nmatch_type\x18\x05 \x01(\x0e\x32\".p4.config.v1.MatchField.MatchTypeH\x00\x12\x1a\n\x10other_match_type\x18\x07 \x01(\tH\x00\x12(\n\x03\x64oc\x18\x06 \x01(\x0b\x32\x1b.p4.config.v1.Documentation\x12,\n\ttype_name\x18\x08 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\x12\x42\n\x16structured_annotations\x18\t \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"V\n\tMatchType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\t\n\x05\x45XACT\x10\x02\x12\x07\n\x03LPM\x10\x03\x12\x0b\n\x07TERNARY\x10\x04\x12\t\n\x05RANGE\x10\x05\x12\x0c\n\x08OPTIONAL\x10\x06\x42\x07\n\x05match\"\xde\x03\n\x05Table\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12.\n\x0cmatch_fields\x18\x02 \x03(\x0b\x32\x18.p4.config.v1.MatchField\x12,\n\x0b\x61\x63tion_refs\x18\x03 \x03(\x0b\x32\x17.p4.config.v1.ActionRef\x12\x1f\n\x17\x63onst_default_action_id\x18\x04 \x01(\r\x12\x19\n\x11implementation_id\x18\x06 \x01(\r\x12\x1b\n\x13\x64irect_resource_ids\x18\x07 \x03(\r\x12\x0c\n\x04size\x18\x08 \x01(\x03\x12\x46\n\x15idle_timeout_behavior\x18\t \x01(\x0e\x32\'.p4.config.v1.Table.IdleTimeoutBehavior\x12\x16\n\x0eis_const_table\x18\n \x01(\x08\x12\x1b\n\x13has_initial_entries\x18\x0b \x01(\x08\x12.\n\x10other_properties\x18\x64 \x01(\x0b\x32\x14.google.protobuf.Any\"9\n\x13IdleTimeoutBehavior\x12\x0e\n\nNO_TIMEOUT\x10\x00\x12\x12\n\x0eNOTIFY_CONTROL\x10\x01\"\x9c\x02\n\tActionRef\x12\n\n\x02id\x18\x01 \x01(\r\x12,\n\x05scope\x18\x03 \x01(\x0e\x32\x1d.p4.config.v1.ActionRef.Scope\x12\x13\n\x0b\x61nnotations\x18\x02 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x05 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x42\n\x16structured_annotations\x18\x04 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"@\n\x05Scope\x12\x15\n\x11TABLE_AND_DEFAULT\x10\x00\x12\x0e\n\nTABLE_ONLY\x10\x01\x12\x10\n\x0c\x44\x45\x46\x41ULT_ONLY\x10\x02\"\x81\x03\n\x06\x41\x63tion\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12*\n\x06params\x18\x02 \x03(\x0b\x32\x1a.p4.config.v1.Action.Param\x1a\xa0\x02\n\x05Param\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x61nnotations\x18\x03 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x08 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x10\n\x08\x62itwidth\x18\x04 \x01(\x05\x12(\n\x03\x64oc\x18\x05 \x01(\x0b\x32\x1b.p4.config.v1.Documentation\x12,\n\ttype_name\x18\x06 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\x12\x42\n\x16structured_annotations\x18\x07 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"\x82\x03\n\rActionProfile\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\x11\n\ttable_ids\x18\x02 \x03(\r\x12\x15\n\rwith_selector\x18\x03 \x01(\x08\x12\x0c\n\x04size\x18\x04 \x01(\x03\x12\x16\n\x0emax_group_size\x18\x05 \x01(\x05\x12\x42\n\x0esum_of_weights\x18\x06 \x01(\x0b\x32(.p4.config.v1.ActionProfile.SumOfWeightsH\x00\x12\x42\n\x0esum_of_members\x18\x07 \x01(\x0b\x32(.p4.config.v1.ActionProfile.SumOfMembersH\x00\x1a\x0e\n\x0cSumOfWeights\x1a\x44\n\x0cSumOfMembers\x12\x1e\n\x11max_member_weight\x18\x01 \x01(\x05H\x00\x88\x01\x01\x42\x14\n\x12_max_member_weightB\x19\n\x17selector_size_semantics\"v\n\x0b\x43ounterSpec\x12,\n\x04unit\x18\x01 \x01(\x0e\x32\x1e.p4.config.v1.CounterSpec.Unit\"9\n\x04Unit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\t\n\x05\x42YTES\x10\x01\x12\x0b\n\x07PACKETS\x10\x02\x12\x08\n\x04\x42OTH\x10\x03\"\x9e\x01\n\x07\x43ounter\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\'\n\x04spec\x18\x02 \x01(\x0b\x32\x19.p4.config.v1.CounterSpec\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x32\n\x0findex_type_name\x18\x04 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\"{\n\rDirectCounter\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\'\n\x04spec\x18\x02 \x01(\x0b\x32\x19.p4.config.v1.CounterSpec\x12\x17\n\x0f\x64irect_table_id\x18\x03 \x01(\r\"h\n\tMeterSpec\x12*\n\x04unit\x18\x01 \x01(\x0e\x32\x1c.p4.config.v1.MeterSpec.Unit\"/\n\x04Unit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\t\n\x05\x42YTES\x10\x01\x12\x0b\n\x07PACKETS\x10\x02\"\x9a\x01\n\x05Meter\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12%\n\x04spec\x18\x02 \x01(\x0b\x32\x17.p4.config.v1.MeterSpec\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x32\n\x0findex_type_name\x18\x04 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\"w\n\x0b\x44irectMeter\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12%\n\x04spec\x18\x02 \x01(\x0b\x32\x17.p4.config.v1.MeterSpec\x12\x17\n\x0f\x64irect_table_id\x18\x03 \x01(\r\"\x83\x03\n\x18\x43ontrollerPacketMetadata\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\x41\n\x08metadata\x18\x02 \x03(\x0b\x32/.p4.config.v1.ControllerPacketMetadata.Metadata\x1a\xf9\x01\n\x08Metadata\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x61nnotations\x18\x03 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x07 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x10\n\x08\x62itwidth\x18\x04 \x01(\x05\x12,\n\ttype_name\x18\x05 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\x12\x42\n\x16structured_annotations\x18\x06 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"k\n\x08ValueSet\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\'\n\x05match\x18\x02 \x03(\x0b\x32\x18.p4.config.v1.MatchField\x12\x0c\n\x04size\x18\x03 \x01(\x05\"\xa7\x01\n\x08Register\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12/\n\ttype_spec\x18\x02 \x01(\x0b\x32\x1c.p4.config.v1.P4DataTypeSpec\x12\x0c\n\x04size\x18\x03 \x01(\x05\x12\x32\n\x0findex_type_name\x18\x04 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\"c\n\x06\x44igest\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12/\n\ttype_spec\x18\x02 \x01(\x0b\x32\x1c.p4.config.v1.P4DataTypeSpecB-Z+github.com/p4lang/p4runtime/go/p4/config/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'p4.config.v1.p4info_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z+github.com/p4lang/p4runtime/go/p4/config/v1'
  _globals['_P4INFO']._serialized_start=99
  _globals['_P4INFO']._serialized_end=747
  _globals['_DOCUMENTATION']._serialized_start=749
  _globals['_DOCUMENTATION']._serialized_end=800
  _globals['_PLATFORMPROPERTIES']._serialized_start=803
  _globals['_PLATFORMPROPERTIES']._serialized_end=959
  _globals['_PKGINFO']._serialized_start=962
  _globals['_PKGINFO']._serialized_end=1322
  _globals['_P4IDS']._serialized_start=1325
  _globals['_P4IDS']._serialized_end=1588
  _globals['_P4IDS_PREFIX']._serialized_start=1335
  _globals['_P4IDS_PREFIX']._serialized_end=1588
  _globals['_PREAMBLE']._serialized_start=1591
  _globals['_PREAMBLE']._serialized_end=1833
  _globals['_EXTERN']._serialized_start=1835
  _globals['_EXTERN']._serialized_end=1942
  _globals['_EXTERNINSTANCE']._serialized_start=1944
  _globals['_EXTERNINSTANCE']._serialized_end=2038
  _globals['_MATCHFIELD']._serialized_start=2041
  _globals['_MATCHFIELD']._serialized_end=2517
  _globals['_MATCHFIELD_MATCHTYPE']._serialized_start=2422
  _globals['_MATCHFIELD_MATCHTYPE']._serialized_end=2508
  _globals['_TABLE']._serialized_start=2520
  _globals['_TABLE']._serialized_end=2998
  _globals['_TABLE_IDLETIMEOUTBEHAVIOR']._serialized_start=2941
  _globals['_TABLE_IDLETIMEOUTBEHAVIOR']._serialized_end=2998
  _globals['_ACTIONREF']._serialized_start=3001
  _globals['_ACTIONREF']._serialized_end=3285
  _globals['_ACTIONREF_SCOPE']._serialized_start=3221
  _globals['_ACTIONREF_SCOPE']._serialized_end=3285
  _globals['_ACTION']._serialized_start=3288
  _globals['_ACTION']._serialized_end=3673
  _globals['_ACTION_PARAM']._serialized_start=3385
  _globals['_ACTION_PARAM']._serialized_end=3673
  _globals['_ACTIONPROFILE']._serialized_start=3676
  _globals['_ACTIONPROFILE']._serialized_end=4062
  _globals['_ACTIONPROFILE_SUMOFWEIGHTS']._serialized_start=3951
  _globals['_ACTIONPROFILE_SUMOFWEIGHTS']._serialized_end=3965
  _globals['_ACTIONPROFILE_SUMOFMEMBERS']._serialized_start=3967
  _globals['_ACTIONPROFILE_SUMOFMEMBERS']._serialized_end=4035
  _globals['_COUNTERSPEC']._serialized_start=4064
  _globals['_COUNTERSPEC']._serialized_end=4182
  _globals['_COUNTERSPEC_UNIT']._serialized_start=4125
  _globals['_COUNTERSPEC_UNIT']._serialized_end=4182
  _globals['_COUNTER']._serialized_start=4185
  _globals['_COUNTER']._serialized_end=4343
  _globals['_DIRECTCOUNTER']._serialized_start=4345
  _globals['_DIRECTCOUNTER']._serialized_end=4468
  _globals['_METERSPEC']._serialized_start=4470
  _globals['_METERSPEC']._serialized_end=4574
  _globals['_METERSPEC_UNIT']._serialized_start=4125
  _globals['_METERSPEC_UNIT']._serialized_end=4172
  _globals['_METER']._serialized_start=4577
  _globals['_METER']._serialized_end=4731
  _globals['_DIRECTMETER']._serialized_start=4733
  _globals['_DIRECTMETER']._serialized_end=4852
  _globals['_CONTROLLERPACKETMETADATA']._serialized_start=4855
  _globals['_CONTROLLERPACKETMETADATA']._serialized_end=5242
  _globals['_CONTROLLERPACKETMETADATA_METADATA']._serialized_start=4993
  _globals['_CONTROLLERPACKETMETADATA_METADATA']._serialized_end=5242
  _globals['_VALUESET']._serialized_start=5244
  _globals['_VALUESET']._serialized_end=5351
  _globals['_REGISTER']._serialized_start=5354
  _globals['_REGISTER']._serialized_end=5521
  _globals['_DIGEST']._serialized_start=5523
  _globals['_DIGEST']._serialized_end=5622
# @@protoc_insertion_point(module_scope)



"Utility functions for opening a GRPC channel."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from dataclasses import dataclass
from pathlib import Path
from typing import Any, cast

import grpc  # pyright: ignore[reportMissingTypeStubs]

from finsy.log import LOGGER
from finsy.proto import rpc_code

# `grpc.aio.EOF` is not typed.
GRPC_EOF: object = cast(
    object, grpc.aio.EOF  # pyright: ignore[reportUnknownMemberType]
)


class _EnumBase(enum.IntEnum):
    "Base class for GRPC/Protobuf Enum wrapper classes."

    def __repr__(self) -> str:
        return f"{type(self).__name__}.{self.name}"


class GRPCStatusCode(_EnumBase):
    "IntEnum equivalent to `grpc.StatusCode`."

    OK = rpc_code.OK
    CANCELLED = rpc_code.CANCELLED
    UNKNOWN = rpc_code.UNKNOWN
    FAILED_PRECONDITION = rpc_code.FAILED_PRECONDITION
    INVALID_ARGUMENT = rpc_code.INVALID_ARGUMENT
    DEADLINE_EXCEEDED = rpc_code.DEADLINE_EXCEEDED
    NOT_FOUND = rpc_code.NOT_FOUND
    ALREADY_EXISTS = rpc_code.ALREADY_EXISTS
    PERMISSION_DENIED = rpc_code.PERMISSION_DENIED
    UNAUTHENTICATED = rpc_code.UNAUTHENTICATED
    RESOURCE_EXHAUSTED = rpc_code.RESOURCE_EXHAUSTED
    ABORTED = rpc_code.ABORTED
    OUT_OF_RANGE = rpc_code.OUT_OF_RANGE
    UNIMPLEMENTED = rpc_code.UNIMPLEMENTED
    INTERNAL = rpc_code.INTERNAL
    UNAVAILABLE = rpc_code.UNAVAILABLE
    DATA_LOSS = rpc_code.DATA_LOSS

    @classmethod
    def from_status_code(cls, val: grpc.StatusCode) -> "GRPCStatusCode":
        "Create corresponding GRPCStatusCode from a grpc.StatusCode object."
        n: Any = val.value[0]  # pyright: ignore[reportUnknownMemberType]
        assert isinstance(n, int)
        return GRPCStatusCode(n)

    @staticmethod
    def _validate_enum() -> None:
        "Verify that GRPCStatusCode covers every possible grpc.StatusCode."
        for value in grpc.StatusCode:
            n: Any = value.value[0]  # pyright: ignore[reportUnknownMemberType]
            assert isinstance(n, int)
            assert GRPCStatusCode[value.name].value == n, value.name


# Check GRPCStatusCode against grpc.StatusCode.
GRPCStatusCode._validate_enum()  # pyright: ignore[reportPrivateUsage]


class GRPCArg(str, enum.Enum):
    """Names for GRPC channel arguments.

    Reference:
    https://grpc.github.io/grpc/core/group__grpc__arg__keys.html
    """

    MAX_METADATA_SIZE = "grpc.max_metadata_size"  # int
    MAX_RECONNECT_BACKOFF_MS = "grpc.max_reconnect_backoff_ms"  # int
    SSL_TARGET_NAME_OVERRIDE_ARG = "grpc.ssl_target_name_override"  # str


GRPCChannelArgumentType = list[tuple[str, int | str]]


@dataclass
class GRPCOptions:
    "GRPC channel arguments."

    max_metadata_size: int | None = None
    max_reconnect_backoff_ms: int | None = None

    def to_native(self) -> GRPCChannelArgumentType:
        "Return GRPC options in form suitable for grpc API."
        results: GRPCChannelArgumentType = []

        if self.max_metadata_size is not None:
            results.append((GRPCArg.MAX_METADATA_SIZE, self.max_metadata_size))
        if self.max_reconnect_backoff_ms is not None:
            results.append(
                (GRPCArg.MAX_RECONNECT_BACKOFF_MS, self.max_reconnect_backoff_ms)
            )
        return results


@dataclass(kw_only=True)
class GRPCCredentialsTLS:
    "GRPC channel credentials for Transport Layer Security (TLS)."

    cacert: Path | bytes | None
    """Certificate authority used to authenticate the certificate at the other
    end of the connection."""

    cert: Path | bytes | None
    "Certificate that identifies this side of the connection."

    private_key: Path | bytes | None
    "Private key associated with this side's certificate identity."

    target_name_override: str = ""
    "Override the target name used for TLS host name checking (useful for testing)."

    call_credentials: grpc.AuthMetadataPlugin | None = None
    """Optional GRPC call credentials for the client channel. Be aware that the
    auth plugin's callback takes place in a different thread."""

    def to_client_credentials(self) -> grpc.ChannelCredentials:
        "Create native SSL client credentials object."
        root_certificates = _coerce_tls_path(self.cacert)
        certificate_chain = _coerce_tls_path(self.cert)
        private_key = _coerce_tls_path(self.private_key)

        return self._compose_credentials(
            grpc.ssl_channel_credentials(  # pyright: ignore[reportUnknownMemberType]
                root_certificates=root_certificates,
                private_key=private_key,
                certificate_chain=certificate_chain,
            )
        )

    def to_server_credentials(self) -> grpc.ServerCredentials:
        """Create native SSL server credentials object.

        On the server side, we ignore the `call_credentials`.
        """
        root_certificates = _coerce_tls_path(self.cacert)
        certificate_chain = _coerce_tls_path(self.cert)
        private_key = _coerce_tls_path(self.private_key)

        return grpc.ssl_server_credentials(  # pyright: ignore[reportUnknownMemberType]
            private_key_certificate_chain_pairs=[(private_key, certificate_chain)],
            root_certificates=root_certificates,
            require_client_auth=True,
        )

    def _compose_credentials(
        self, channel_cred: grpc.ChannelCredentials
    ) -> grpc.ChannelCredentials:
        "Compose call credentials with channel credentials."
        if not self.call_credentials:
            return channel_cred

        call_cred = (
            grpc.metadata_call_credentials(  # pyright: ignore[reportUnknownMemberType]
                self.call_credentials
            )
        )
        return grpc.composite_channel_credentials(  # pyright: ignore[reportUnknownMemberType]
            channel_cred,
            call_cred,
        )


def _coerce_tls_path(value: Path | bytes | None) -> bytes | None:
    "Convert Path to bytes."
    if isinstance(value, Path):
        return value.read_bytes()
    assert value is None or isinstance(value, bytes)
    return value


def grpc_channel(
    address: str,
    *,
    credentials: GRPCCredentialsTLS | None = None,
    options: GRPCOptions | None = None,
    client_type: str = "GRPC",
) -> grpc.aio.Channel:
    "Create a GRPC AIO channel."
    if credentials is None:
        LOGGER.debug("%s: create INSECURE channel %r", client_type, address)
        channel = grpc.aio.insecure_channel(
            address,
            options=options.to_native() if options else None,
        )
    else:
        LOGGER.debug("%s: create secure channel %r", client_type, address)
        args: GRPCChannelArgumentType = options.to_native() if options else []
        if credentials.target_name_override:
            args.append(
                (GRPCArg.SSL_TARGET_NAME_OVERRIDE_ARG, credentials.target_name_override)
            )
        channel = grpc.aio.secure_channel(
            address,
            credentials=credentials.to_client_credentials(),
            options=args,
        )

    return channel



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import p4runtime_pb2 as p4_dot_v1_dot_p4runtime__pb2

GRPC_GENERATED_VERSION = '1.64.0'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in p4/v1/p4runtime_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class P4RuntimeStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Write = channel.unary_unary(
                '/p4.v1.P4Runtime/Write',
                request_serializer=p4_dot_v1_dot_p4runtime__pb2.WriteRequest.SerializeToString,
                response_deserializer=p4_dot_v1_dot_p4runtime__pb2.WriteResponse.FromString,
                _registered_method=True)
        self.Read = channel.unary_stream(
                '/p4.v1.P4Runtime/Read',
                request_serializer=p4_dot_v1_dot_p4runtime__pb2.ReadRequest.SerializeToString,
                response_deserializer=p4_dot_v1_dot_p4runtime__pb2.ReadResponse.FromString,
                _registered_method=True)
        self.SetForwardingPipelineConfig = channel.unary_unary(
                '/p4.v1.P4Runtime/SetForwardingPipelineConfig',
                request_serializer=p4_dot_v1_dot_p4runtime__pb2.SetForwardingPipelineConfigRequest.SerializeToString,
                response_deserializer=p4_dot_v1_dot_p4runtime__pb2.SetForwardingPipelineConfigResponse.FromString,
                _registered_method=True)
        self.GetForwardingPipelineConfig = channel.unary_unary(
                '/p4.v1.P4Runtime/GetForwardingPipelineConfig',
                request_serializer=p4_dot_v1_dot_p4runtime__pb2.GetForwardingPipelineConfigRequest.SerializeToString,
                response_deserializer=p4_dot_v1_dot_p4runtime__pb2.GetForwardingPipelineConfigResponse.FromString,
                _registered_method=True)
        self.StreamChannel = channel.stream_stream(
                '/p4.v1.P4Runtime/StreamChannel',
                request_serializer=p4_dot_v1_dot_p4runtime__pb2.StreamMessageRequest.SerializeToString,
                response_deserializer=p4_dot_v1_dot_p4runtime__pb2.StreamMessageResponse.FromString,
                _registered_method=True)
        self.Capabilities = channel.unary_unary(
                '/p4.v1.P4Runtime/Capabilities',
                request_serializer=p4_dot_v1_dot_p4runtime__pb2.CapabilitiesRequest.SerializeToString,
                response_deserializer=p4_dot_v1_dot_p4runtime__pb2.CapabilitiesResponse.FromString,
                _registered_method=True)


class P4RuntimeServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Write(self, request, context):
        """Update one or more P4 entities on the target.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Read(self, request, context):
        """Read one or more P4 entities from the target.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetForwardingPipelineConfig(self, request, context):
        """Sets the P4 forwarding-pipeline config.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetForwardingPipelineConfig(self, request, context):
        """Gets the current P4 forwarding-pipeline config.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StreamChannel(self, request_iterator, context):
        """Represents the bidirectional stream between the controller and the
        switch (initiated by the controller), and is managed for the following
        purposes:
        - connection initiation through client arbitration
        - indicating switch session liveness: the session is live when switch
        sends a positive client arbitration update to the controller, and is
        considered dead when either the stream breaks or the switch sends a
        negative update for client arbitration
        - the controller sending/receiving packets to/from the switch
        - streaming of notifications from the switch
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Capabilities(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_P4RuntimeServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Write': grpc.unary_unary_rpc_method_handler(
                    servicer.Write,
                    request_deserializer=p4_dot_v1_dot_p4runtime__pb2.WriteRequest.FromString,
                    response_serializer=p4_dot_v1_dot_p4runtime__pb2.WriteResponse.SerializeToString,
            ),
            'Read': grpc.unary_stream_rpc_method_handler(
                    servicer.Read,
                    request_deserializer=p4_dot_v1_dot_p4runtime__pb2.ReadRequest.FromString,
                    response_serializer=p4_dot_v1_dot_p4runtime__pb2.ReadResponse.SerializeToString,
            ),
            'SetForwardingPipelineConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.SetForwardingPipelineConfig,
                    request_deserializer=p4_dot_v1_dot_p4runtime__pb2.SetForwardingPipelineConfigRequest.FromString,
                    response_serializer=p4_dot_v1_dot_p4runtime__pb2.SetForwardingPipelineConfigResponse.SerializeToString,
            ),
            'GetForwardingPipelineConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.GetForwardingPipelineConfig,
                    request_deserializer=p4_dot_v1_dot_p4runtime__pb2.GetForwardingPipelineConfigRequest.FromString,
                    response_serializer=p4_dot_v1_dot_p4runtime__pb2.GetForwardingPipelineConfigResponse.SerializeToString,
            ),
            'StreamChannel': grpc.stream_stream_rpc_method_handler(
                    servicer.StreamChannel,
                    request_deserializer=p4_dot_v1_dot_p4runtime__pb2.StreamMessageRequest.FromString,
                    response_serializer=p4_dot_v1_dot_p4runtime__pb2.StreamMessageResponse.SerializeToString,
            ),
            'Capabilities': grpc.unary_unary_rpc_method_handler(
                    servicer.Capabilities,
                    request_deserializer=p4_dot_v1_dot_p4runtime__pb2.CapabilitiesRequest.FromString,
                    response_serializer=p4_dot_v1_dot_p4runtime__pb2.CapabilitiesResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'p4.v1.P4Runtime', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('p4.v1.P4Runtime', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class P4Runtime(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Write(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/p4.v1.P4Runtime/Write',
            p4_dot_v1_dot_p4runtime__pb2.WriteRequest.SerializeToString,
            p4_dot_v1_dot_p4runtime__pb2.WriteResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Read(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/p4.v1.P4Runtime/Read',
            p4_dot_v1_dot_p4runtime__pb2.ReadRequest.SerializeToString,
            p4_dot_v1_dot_p4runtime__pb2.ReadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetForwardingPipelineConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/p4.v1.P4Runtime/SetForwardingPipelineConfig',
            p4_dot_v1_dot_p4runtime__pb2.SetForwardingPipelineConfigRequest.SerializeToString,
            p4_dot_v1_dot_p4runtime__pb2.SetForwardingPipelineConfigResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetForwardingPipelineConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/p4.v1.P4Runtime/GetForwardingPipelineConfig',
            p4_dot_v1_dot_p4runtime__pb2.GetForwardingPipelineConfigRequest.SerializeToString,
            p4_dot_v1_dot_p4runtime__pb2.GetForwardingPipelineConfigResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StreamChannel(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/p4.v1.P4Runtime/StreamChannel',
            p4_dot_v1_dot_p4runtime__pb2.StreamMessageRequest.SerializeToString,
            p4_dot_v1_dot_p4runtime__pb2.StreamMessageResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Capabilities(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/p4.v1.P4Runtime/Capabilities',
            p4_dot_v1_dot_p4runtime__pb2.CapabilitiesRequest.SerializeToString,
            p4_dot_v1_dot_p4runtime__pb2.CapabilitiesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"Implements the Switch class."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import dataclasses
import enum
import logging
import re
import time
from asyncio import Queue
from contextlib import asynccontextmanager
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Callable,
    Coroutine,
    Iterable,
    NamedTuple,
    SupportsBytes,
    TypeVar,
    final,
    overload,
)

import pyee
from typing_extensions import Self

from finsy import p4entity, pbutil
from finsy.futures import CountdownFuture
from finsy.gnmiclient import GNMIClient, GNMIClientError
from finsy.grpcutil import GRPCCredentialsTLS, GRPCStatusCode
from finsy.log import LOGGER
from finsy.p4arbitrator import Arbitrator
from finsy.p4client import P4Client, P4ClientError
from finsy.p4schema import P4ConfigAction, P4ConfigResponseType, P4Schema
from finsy.ports import SwitchPortList
from finsy.proto import p4r

# Maximum size of queues used for PacketIn, DigestList, etc.

_DEFAULT_QUEUE_SIZE = 50

# If switch fails to run for longer than _FAIL_MIN_TIME_SECS, wait
# _FAIL_THROTTLE_SECS before trying again.

_FAIL_MIN_TIME_SECS = 2.0
_FAIL_THROTTLE_SECS = 10.0

_T = TypeVar("_T")
_ET = TypeVar("_ET", bound=p4entity.P4Entity)


@final
@dataclasses.dataclass(frozen=True)
class SwitchOptions:
    """Represents the configuration options for a `Switch`.

    ```
    opts = SwitchOptions(
        p4info=Path("basic.p4info.txtpb"),
        p4blob=Path("basic.json"),
        ready_handler=on_ready,
    )
    ```

    Each `SwitchOptions` object is immutable and may be shared by multiple
    switches. You should treat all values as read-only.

    You can use function call syntax to return a copy of a `SwitchOptions` with
    one or more propertise altered.

    ```
    new_opts = opts(device_id=6)
    ```
    """

    p4info: Path | None = None
    "Path to P4Info protobuf text file."

    p4blob: Path | SupportsBytes | None = None
    "Path to P4Blob file, or an object that can provide the bytes value."

    p4force: bool = False
    "If true, always load the P4 program after initial handshake."

    device_id: int = 1
    "Default P4Runtime device ID."

    initial_election_id: int = 10
    "Initial P4Runtime election ID."

    channel_credentials: GRPCCredentialsTLS | None = None
    "P4Runtime channel credentials. Used for TLS support."

    role_name: str = ""
    "P4Runtime role configuration."

    role_config: pbutil.PBMessage | None = None
    "P4Runtime role configuration."

    ready_handler: Callable[["Switch"], Coroutine[Any, Any, None]] | None = None
    "Ready handler async function callback."

    fail_fast: bool = False
    """If true, log switch errors as CRITICAL and immediately abort when the
    switch is running in a Controller."""

    def __call__(self, **kwds: Any) -> Self:
        return dataclasses.replace(self, **kwds)


_SEMVER_REGEX = re.compile(r"^(\d+)\.(\d+)(?:\.(\d+))?(.*)$")


class ApiVersion(NamedTuple):
    "Represents the semantic version of an API."

    major: int
    minor: int
    patch: int
    extra: str  # optional pre-release/build info

    @classmethod
    def parse(cls, version_str: str) -> Self:
        "Parse the API version string."
        m = _SEMVER_REGEX.match(version_str.strip())
        if not m:
            raise ValueError(f"unexpected version string: {version_str}")
        return cls(int(m[1]), int(m[2]), int(m[3] or "0"), m[4])

    def __str__(self) -> str:
        "Return the version string."
        vers = self[:3]  # pylint: disable=unsubscriptable-object
        return ".".join(map(str, vers)) + self.extra


@final
class Switch:
    """Represents a P4Runtime Switch.

    A `Switch` is constructed with a `name`, `address` and an optional
    `SwitchOptions` configuration.

    The `name` is up to the user but should uniquely identify the switch.

    The `address` identifies the target endpoint of the GRPC channel. It should
    have the format "<address>:<port>" where <address> can be a domain name,
    IPv4 address, or IPv6 address in square brackets, and <port> is the TCP
    port number.

    The `options` is a `SwitchOptions` object that specifies how the `Switch`
    will behave.

    ```
    opts = SwitchOptions(p4info=..., p4blob=...)
    sw1 = Switch('sw1', '10.0.0.1:50000', opts)
    ```

    The `stash` is an optional dictionary for storing arbitrary per-switch
    information. The dictionary keys are strings. You can use this to store
    anything that you need to manage the switch.

    Each switch object has an event emitter `ee`. Use the EventEmitter to listen
    for port change events like PORT_UP and PORT_DOWN. See the `SwitchEvent`
    class for a list of other switch events.
    """

    _name: str
    _address: str
    _options: SwitchOptions
    _stash: dict[str, Any]
    _ee: "SwitchEmitter"
    _p4client: P4Client | None
    _p4schema: P4Schema
    _tasks: "SwitchTasks | None"
    _packet_queues: list[tuple[Callable[[bytes], bool], Queue[p4entity.P4PacketIn]]]
    _digest_queues: dict[str, Queue[p4entity.P4DigestList]]
    _timeout_queue: Queue[p4entity.P4IdleTimeoutNotification] | None
    _arbitrator: "Arbitrator"
    _gnmi_client: GNMIClient | None
    _ports: SwitchPortList
    _is_channel_up: bool = False
    _api_version: ApiVersion = ApiVersion(1, 0, 0, "")
    _control_task: asyncio.Task[Any] | None = None

    def __init__(
        self,
        name: str,
        address: str,
        options: SwitchOptions | None = None,
        *,
        stash: dict[str, Any] | None = None,
    ) -> None:
        """Initialize switch with name, address and options.

        Args:
            name: A human-readable name to uniquely identify the switch.
            address: The target address of the P4Runtime GRPC channel.
              Format is "<address>:<port>" where <address> is a DNS name or
              IP address, and <port> is the TCP port number.
            options: Configuration options for the switch.
            stash: Optional user-controlled dictionary.
              Used to store information that user code needs to access or share.
        """
        if options is None:
            options = SwitchOptions()

        self._name = name
        self._address = address
        self._options = options
        self._stash = stash or {}
        self._ee = SwitchEmitter(self)
        self._p4client = None
        self._p4schema = P4Schema(options.p4info, options.p4blob)
        self._tasks = None
        self._packet_queues = []
        self._digest_queues = {}
        self._timeout_queue = None
        self._arbitrator = Arbitrator(
            options.initial_election_id, options.role_name, options.role_config
        )
        self._gnmi_client = None
        self._ports = SwitchPortList()

    @property
    def name(self) -> str:
        "Name of the switch."
        return self._name

    @property
    def address(self) -> str:
        "Address of the switch."
        return self._address

    @property
    def options(self) -> SwitchOptions:
        "Switch options."
        return self._options

    @options.setter
    def options(self, opts: SwitchOptions) -> None:
        "Set switch options to a new value."
        if self._p4client is not None:
            raise RuntimeError("Cannot change switch options while client is open.")

        self._options = opts
        self._p4schema = P4Schema(opts.p4info, opts.p4blob)
        self._arbitrator = Arbitrator(
            opts.initial_election_id, opts.role_name, opts.role_config
        )

    @property
    def stash(self) -> dict[str, Any]:
        "Switch stash. Used to store per-switch data for any purpose."
        return self._stash

    @property
    def ee(self) -> "SwitchEmitter":
        "Switch event emitter. See `SwitchEvent` for more details on events."
        return self._ee

    @property
    def device_id(self) -> int:
        "Switch's device ID."
        return self._options.device_id

    @property
    def is_up(self) -> bool:
        "True if switch is UP."
        return self._is_channel_up

    @property
    def is_primary(self) -> bool:
        "True if switch is primary."
        return self._arbitrator.is_primary

    @property
    def primary_id(self) -> int:
        "Election ID of switch that is currently primary."
        return self._arbitrator.primary_id

    @property
    def election_id(self) -> int:
        "Switch's current election ID."
        return self._arbitrator.election_id

    @property
    def role_name(self) -> str:
        "Switch's current role name."
        return self._arbitrator.role_name

    @property
    def p4info(self) -> P4Schema:
        "Switch's P4 schema."
        return self._p4schema

    @property
    def gnmi_client(self) -> GNMIClient | None:
        "Switch's gNMI client."
        return self._gnmi_client

    @property
    def ports(self) -> SwitchPortList:
        "Switch's list of interfaces."
        return self._ports

    @property
    def api_version(self) -> ApiVersion:
        "P4Runtime protocol version."
        return self._api_version

    @overload
    async def read(
        self,
        entities: _ET,
    ) -> AsyncGenerator[_ET, None]:
        "Overload for read of a single P4Entity subtype."
        ...  # pragma: no cover

    @overload
    async def read(
        self,
        entities: Iterable[_ET],
    ) -> AsyncGenerator[_ET, None]:
        "Overload for read of an iterable of the same P4Entity subtype."
        ...  # pragma: no cover

    @overload
    async def read(
        self,
        entities: Iterable[p4entity.P4EntityList],
    ) -> AsyncGenerator[p4entity.P4Entity, None]:
        "Most general overload: we can't determine the return type exactly."
        ...  # pragma: no cover

    async def read(
        self,
        entities: Iterable[p4entity.P4EntityList] | p4entity.P4Entity,
    ) -> AsyncGenerator[p4entity.P4Entity, None]:
        "Async iterator that reads entities from the switch."
        assert self._p4client is not None

        if not entities:
            return

        if isinstance(entities, p4entity.P4Entity):
            entities = [entities]

        request = p4r.ReadRequest(
            device_id=self.device_id,
            entities=p4entity.encode_entities(entities, self.p4info),
        )

        async for reply in self._p4client.request_iter(request):
            for ent in reply.entities:
                yield p4entity.decode_entity(ent, self.p4info)

    async def read_packets(
        self,
        *,
        queue_size: int = _DEFAULT_QUEUE_SIZE,
        eth_types: Iterable[int] | None = None,
    ) -> AsyncIterator["p4entity.P4PacketIn"]:
        "Async iterator for incoming packets (P4PacketIn)."
        LOGGER.debug("read_packets: opening queue: eth_types=%r", eth_types)

        if eth_types is None:

            def _pkt_filter(_payload: bytes) -> bool:
                return True

        else:
            _filter = {eth.to_bytes(2, "big") for eth in eth_types}

            def _pkt_filter(_payload: bytes) -> bool:
                return _payload[12:14] in _filter

        queue = Queue[p4entity.P4PacketIn](queue_size)
        queue_filter = (_pkt_filter, queue)
        self._packet_queues.append(queue_filter)

        try:
            while True:
                yield await queue.get()
        finally:
            LOGGER.debug("read_packets: closing queue: eth_types=%r", eth_types)
            self._packet_queues.remove(queue_filter)

    async def read_digests(
        self,
        digest_id: str,
        *,
        queue_size: int = _DEFAULT_QUEUE_SIZE,
    ) -> AsyncIterator["p4entity.P4DigestList"]:
        "Async iterator for incoming digest lists (P4DigestList)."
        LOGGER.debug("read_digests: opening queue: digest_id=%r", digest_id)

        if digest_id in self._digest_queues:
            raise ValueError(f"queue for digest_id {digest_id!r} already open")

        queue = Queue[p4entity.P4DigestList](queue_size)
        self._digest_queues[digest_id] = queue
        try:
            while True:
                yield await queue.get()
        finally:
            LOGGER.debug("read_digests: closing queue: digest_id=%r", digest_id)
            del self._digest_queues[digest_id]

    async def read_idle_timeouts(
        self,
        *,
        queue_size: int = _DEFAULT_QUEUE_SIZE,
    ) -> AsyncIterator["p4entity.P4IdleTimeoutNotification"]:
        "Async iterator for incoming idle timeouts (P4IdleTimeoutNotification)."
        LOGGER.debug("read_idle_timeouts: opening queue")

        if self._timeout_queue is not None:
            raise ValueError("timeout queue already open")

        queue = Queue[p4entity.P4IdleTimeoutNotification](queue_size)
        self._timeout_queue = queue
        try:
            while True:
                yield await queue.get()
        finally:
            LOGGER.debug("read_idle_timeouts: closing queue")
            self._timeout_queue = None

    async def write(
        self,
        entities: Iterable[p4entity.P4UpdateList],
        *,
        strict: bool = True,
        warn_only: bool = False,
    ) -> None:
        """Write updates and stream messages to the switch.

        If `strict` is False, MODIFY and DELETE operations will NOT raise an
        error if the entity does not exist (NOT_FOUND).

        If `warn_only` is True, no operations will raise an error. Instead,
        the exception will be logged as a WARNING and the method will return
        normally.
        """
        assert self._p4client is not None

        if not entities:
            return

        msgs = p4entity.encode_updates(entities, self.p4info)

        updates: list[p4r.Update] = []
        for msg in msgs:
            if isinstance(msg, p4r.StreamMessageRequest):
                # StreamMessageRequests are transmitted immediately.
                # TODO: Understand what happens with backpressure?
                await self._p4client.send(msg)
            else:
                updates.append(msg)

        if updates:
            await self._write_request(updates, strict, warn_only)

    async def insert(
        self,
        entities: Iterable[p4entity.P4EntityList],
        *,
        warn_only: bool = False,
    ) -> None:
        """Insert the specified entities.

        If `warn_only` is True, errors will be logged as warnings instead of
        raising an exception.
        """
        if entities:
            await self._write_request(
                [
                    p4r.Update(type=p4r.Update.INSERT, entity=ent)
                    for ent in p4entity.encode_entities(entities, self.p4info)
                ],
                True,
                warn_only,
            )

    async def modify(
        self,
        entities: Iterable[p4entity.P4EntityList],
        *,
        strict: bool = True,
        warn_only: bool = False,
    ) -> None:
        """Modify the specified entities.

        If `strict` is False, NOT_FOUND errors will be ignored.

        If `warn_only` is True, errors will be logged as warnings instead of
        raising an exception.
        """
        if entities:
            await self._write_request(
                [
                    p4r.Update(type=p4r.Update.MODIFY, entity=ent)
                    for ent in p4entity.encode_entities(entities, self.p4info)
                ],
                strict,
                warn_only,
            )

    async def delete(
        self,
        entities: Iterable[p4entity.P4EntityList],
        *,
        strict: bool = True,
        warn_only: bool = False,
    ) -> None:
        """Delete the specified entities.

        If `strict` is False, NOT_FOUND errors will be ignored.

        If `warn_only` is True, errors will be logged as warnings instead of
        raising an exception.
        """
        if entities:
            await self._write_request(
                [
                    p4r.Update(type=p4r.Update.DELETE, entity=ent)
                    for ent in p4entity.encode_entities(entities, self.p4info)
                ],
                strict,
                warn_only,
            )

    async def delete_all(self) -> None:
        """Delete all entities if no parameter is passed. Otherwise, delete
        items that match `entities`.

        This method does not attempt to delete entries in const tables.

        TODO: This method does not affect indirect counters, meters or
        value_sets.
        """
        await self.delete_many(
            [
                p4entity.P4TableEntry(),
                p4entity.P4MulticastGroupEntry(),
                p4entity.P4CloneSessionEntry(),
            ]
        )

        # Reset all default table entries.
        default_entries = [
            p4entity.P4TableEntry(table.alias, is_default_action=True)
            for table in self.p4info.tables
            if table.const_default_action is None and table.action_profile is None
        ]
        if default_entries:
            await self.modify(default_entries)

        # Delete all P4ActionProfileGroup's and P4ActionProfileMember's.
        # We do this after deleting the P4TableEntry's in case a client is using
        # "one-shot" references; these are incompatible with separate
        # action profiles.
        await self.delete_many(
            [
                p4entity.P4ActionProfileGroup(),
                p4entity.P4ActionProfileMember(),
            ]
        )

        # Delete DigestEntry separately. Wildcard reads are not supported.
        digest_entries = [
            p4entity.P4DigestEntry(digest.alias) for digest in self.p4info.digests
        ]
        if digest_entries:
            await self.delete(digest_entries, strict=False)

    async def delete_many(self, entities: Iterable[p4entity.P4EntityList]) -> None:
        """Delete entities that match a wildcard read.

        This method always skips over entries in const tables. It is an error
        to attempt to delete those.
        """
        assert self._p4client is not None

        request = p4r.ReadRequest(
            device_id=self.device_id,
            entities=p4entity.encode_entities(entities, self.p4info),
        )

        # Compute set of all const table ID's (may be empty).
        to_skip = {table.id for table in self.p4info.tables if table.is_const}

        async for reply in self._p4client.request_iter(request):
            if reply.entities:
                if to_skip:
                    await self.delete(
                        reply
                        for reply in reply.entities
                        if reply.HasField("table_entry")
                        and reply.table_entry.table_id not in to_skip
                    )
                else:
                    await self.delete(reply.entities)

    async def run(self) -> None:
        "Run the switch's lifecycle repeatedly."
        assert self._p4client is None
        assert self._tasks is None

        self._tasks = SwitchTasks(self._options.fail_fast)
        self._p4client = P4Client(self._address, self._options.channel_credentials)
        self._switch_start()

        try:
            while True:
                # If the switch fails and restarts too quickly, slow it down.
                async with _throttle_failure():
                    self.create_task(self._run(), background=True)
                    await self._tasks.wait()
                    self._arbitrator.reset()

        finally:
            self._p4client = None
            self._tasks = None
            self._switch_stop()

    def create_task(
        self,
        coro: Coroutine[Any, Any, _T],
        *,
        background: bool = False,
        name: str | None = None,
    ) -> asyncio.Task[_T]:
        "Create an asyncio task tied to the Switch's lifecycle."
        assert self._tasks is not None

        return self._tasks.create_task(
            coro,
            switch=self,
            background=background,
            name=name,
        )

    async def _run(self):
        "Main Switch task runs the stream."
        assert not self._is_channel_up
        assert self._p4client is not None

        try:
            await self._p4client.open(
                schema=self.p4info,
                complete_request=self._arbitrator.complete_request,
            )
            await self._arbitrator.handshake(self)
            await self._fetch_capabilities()
            await self._start_gnmi()
            self._channel_up()

            # Receive messages from the stream until it closes.
            await self._receive_until_closed()

        finally:
            await self._stop_gnmi()
            await self._p4client.close()
            self._channel_down()

    async def _receive_until_closed(self):
        "Receive messages from stream until EOF."
        assert self._p4client is not None

        client = self._p4client

        while True:
            try:
                msg = await client.receive()
            except P4ClientError as ex:
                if not ex.is_election_id_used:
                    raise
                # Handle "election ID in use" error.
                await self._arbitrator.handshake(self, conflict=True)
            else:
                await self._handle_stream_message(msg)

    async def _handle_stream_message(self, msg: p4r.StreamMessageResponse):
        "Handle a P4Runtime StreamMessageResponse."
        match msg.WhichOneof("update"):
            case "packet":
                self._stream_packet_message(msg)
            case "digest":
                self._stream_digest_message(msg)
            case "idle_timeout_notification":
                self._stream_timeout_message(msg)
            case "arbitration":
                await self._arbitrator.update(self, msg.arbitration)
            case "error":
                self._stream_error_message(msg)
            case other:
                LOGGER.error("_handle_stream_message: unknown update %r", other)

    async def __aenter__(self) -> Self:
        "Similar to run() but provides a one-time context manager interface."
        assert self._p4client is None
        assert self._tasks is None

        self._tasks = SwitchTasks(self._options.fail_fast)
        self._p4client = P4Client(
            self._address,
            self._options.channel_credentials,
            wait_for_ready=False,
        )
        self._switch_start()

        try:
            # Start the switch's `_run` task in the background. Then, wait for
            # `_run` task to fire the CHANNEL_READY event. If the `_run` task
            # cannot connect or fails in some other way, it will finish before
            # the `ready` future. We need to handle the error in this case.

            run = self.create_task(self._run(), background=True)
            ready = self.ee.event_future(SwitchEvent.CHANNEL_READY)
            done, _ = await asyncio.wait(
                [run, ready], return_when=asyncio.FIRST_COMPLETED
            )
            if run in done:
                await run

        except BaseException:
            await self.__aexit__(None, None, None)
            raise

        return self

    async def __aexit__(
        self,
        _exc_type: type[BaseException] | None,
        _exc_val: BaseException | None,
        _exc_tb: TracebackType | None,
    ) -> bool | None:
        "Similar to run() but provides a one-time context manager interface."
        assert self._p4client is not None
        assert self._tasks is not None

        self._tasks.cancel_all()
        await self._tasks.wait()
        self._arbitrator.reset()
        self._p4client = None
        self._tasks = None
        self._switch_stop()

    def _switch_start(self):
        "Called when switch starts its run() cycle."
        assert not self._is_channel_up

        LOGGER.info(
            "Switch start (name=%r, address=%r, device_id=%r, role_name=%r, initial_election_id=%r)",
            self.name,
            self.address,
            self.device_id,
            self.role_name,
            self.options.initial_election_id,
        )
        self.ee.emit(SwitchEvent.SWITCH_START)

    def _switch_stop(self):
        "Called when switch stops its run() cycle."
        assert not self._is_channel_up

        LOGGER.info(
            "Switch stop (name=%r, address=%r, device_id=%r, role_name=%r)",
            self.name,
            self.address,
            self.device_id,
            self.role_name,
        )
        self.ee.emit(SwitchEvent.SWITCH_STOP)

    def _channel_up(self):
        "Called when P4Runtime channel is UP."
        assert not self._is_channel_up

        ports = " ".join(f"({port.id}){port.name}" for port in self.ports)
        LOGGER.info(
            "Channel up (is_primary=%r, role_name=%r, p4r=%s): %s",
            self.is_primary,
            self.role_name,
            self.api_version,
            ports,
        )
        self._is_channel_up = True
        self.create_task(self._ready())

        self.ee.emit(SwitchEvent.CHANNEL_UP, self)

    def _channel_down(self):
        "Called when P4Runtime channel is DOWN."
        if not self._is_channel_up:
            return  # do nothing!

        LOGGER.info(
            "Channel down (is_primary=%r, role_name=%r)",
            self.is_primary,
            self.role_name,
        )
        self._is_channel_up = False

        self.ee.emit(SwitchEvent.CHANNEL_DOWN, self)

    def _become_primary(self):
        "Called when a P4Runtime backup channel becomes the primary."
        assert self._tasks is not None

        LOGGER.info(
            "Become primary (is_primary=%r, role_name=%r)",
            self.is_primary,
            self.role_name,
        )

        self._tasks.cancel_primary()
        self.create_task(self._ready())

        self.ee.emit(SwitchEvent.BECOME_PRIMARY, self)

    def _become_backup(self):
        "Called when a P4Runtime primary channel becomes a backup."
        assert self._tasks is not None

        LOGGER.info(
            "Become backup (is_primary=%r, role_name=%r)",
            self.is_primary,
            self.role_name,
        )

        self._tasks.cancel_primary()
        self.create_task(self._ready())

        self.ee.emit(SwitchEvent.BECOME_BACKUP, self)

    def _channel_ready(self):
        "Called when a P4Runtime channel is READY."
        LOGGER.info(
            "Channel ready (is_primary=%r, role_name=%r): %s",
            self.is_primary,
            self.role_name,
            self.p4info.get_pipeline_info(),
        )

        if self._options.ready_handler:
            self.create_task(self._options.ready_handler(self))

        self.ee.emit(SwitchEvent.CHANNEL_READY, self)

    def _stream_packet_message(self, msg: p4r.StreamMessageResponse):
        "Called when a P4Runtime packet-in response is received."
        packet = p4entity.decode_stream(msg, self.p4info)

        was_queued = False
        for pkt_filter, queue in self._packet_queues:
            if not queue.full() and pkt_filter(packet.payload):
                queue.put_nowait(packet)
                was_queued = True

        if not was_queued:
            LOGGER.warning("packet ignored: %r", packet)

    def _stream_digest_message(self, msg: p4r.StreamMessageResponse):
        "Called when a P4Runtime digest response is received."
        try:
            # Decode the digest list message.
            digest: p4entity.P4DigestList = p4entity.decode_stream(msg, self.p4info)
        except ValueError as ex:
            # It's possible to receive a digest for a different P4Info file, or
            # even before a P4Info is fetched from the switch.
            LOGGER.warning("digest decode failed: %s", ex)
        else:
            # Place the decoded digest list in a queue, if one is waiting.
            queue = self._digest_queues.get(digest.digest_id)
            if queue is not None and not queue.full():
                queue.put_nowait(digest)
            else:
                LOGGER.warning("digest ignored: %r", digest)

    def _stream_timeout_message(self, msg: p4r.StreamMessageResponse):
        "Called when a P4Runtime timeout notification is received."
        timeout: p4entity.P4IdleTimeoutNotification = p4entity.decode_stream(
            msg, self.p4info
        )
        queue = self._timeout_queue

        if queue is not None and not queue.full():
            queue.put_nowait(timeout)
        else:
            LOGGER.warning("timeout ignored: %r", timeout)

    def _stream_error_message(self, msg: p4r.StreamMessageResponse):
        "Called when a P4Runtime stream error response is received."
        assert self._p4client is not None

        # Log the message at the ERROR level.
        pbutil.log_msg(self._p4client.channel, msg, self.p4info, level=logging.ERROR)

        self.ee.emit(SwitchEvent.STREAM_ERROR, self, msg)

    async def _ready(self):
        "Prepare the pipeline."
        if self.p4info.is_authoritative and self.is_primary:
            await self._set_pipeline()
        else:
            await self._get_pipeline()

        self._channel_ready()

    async def _get_pipeline(self):
        "Get the switch's P4Info."
        has_pipeline = False

        try:
            reply = await self._get_pipeline_config_request(
                response_type=P4ConfigResponseType.P4INFO_AND_COOKIE
            )

            if reply.config.HasField("p4info"):
                has_pipeline = True
                p4info = reply.config.p4info
                if not self.p4info.exists:
                    # If we don't have P4Info yet, set it.
                    self.p4info.set_p4info(p4info)
                elif not self.p4info.has_p4info(p4info):
                    # If P4Info is not identical, log a warning message.
                    LOGGER.warning("Retrieved P4Info is different than expected!")

        except P4ClientError as ex:
            if not ex.is_pipeline_missing:
                raise

        if not has_pipeline and self.p4info.exists:
            LOGGER.warning("Forwarding pipeline is not configured")

    async def _set_pipeline(self):
        """Set up the pipeline.

        If `p4force` is false (the default), we first retrieve the cookie for
        the current pipeline and see if it matches the new pipeline's cookie.
        If the cookies match, we are done; there is no need to set the pipeline.

        If `p4force` is true, we always load the new pipeline.
        """
        cookie = -1
        try:
            if not self.options.p4force:
                reply = await self._get_pipeline_config_request()
                if reply.config.HasField("cookie"):
                    cookie = reply.config.cookie.cookie

        except P4ClientError as ex:
            if not ex.is_pipeline_missing:
                raise

        if cookie != self.p4info.p4cookie:
            LOGGER.debug(
                "cookie %#x does not match expected %#x", cookie, self.p4info.p4cookie
            )
            await self._set_pipeline_config_request(
                config=self.p4info.get_pipeline_config()
            )
            LOGGER.info("Pipeline installed: %s", self.p4info.get_pipeline_info())

    async def _get_pipeline_config_request(
        self,
        *,
        response_type: P4ConfigResponseType = P4ConfigResponseType.COOKIE_ONLY,
    ) -> p4r.GetForwardingPipelineConfigResponse:
        "Send a GetForwardingPipelineConfigRequest and await the response."
        assert self._p4client is not None

        return await self._p4client.request(
            p4r.GetForwardingPipelineConfigRequest(
                device_id=self.device_id,
                response_type=response_type.vt(),
            )
        )

    async def _set_pipeline_config_request(
        self,
        *,
        action: P4ConfigAction = P4ConfigAction.VERIFY_AND_COMMIT,
        config: p4r.ForwardingPipelineConfig,
    ) -> p4r.SetForwardingPipelineConfigResponse:
        "Send a SetForwardingPipelineConfigRequest and await the response."
        assert self._p4client is not None

        return await self._p4client.request(
            p4r.SetForwardingPipelineConfigRequest(
                device_id=self.device_id,
                action=action.vt(),
                config=config,
            )
        )

    async def _write_request(
        self,
        updates: list[p4r.Update],
        strict: bool,
        warn_only: bool,
    ):
        "Send a P4Runtime WriteRequest."
        assert self._p4client is not None

        try:
            await self._p4client.request(
                p4r.WriteRequest(
                    device_id=self.device_id,
                    updates=updates,
                )
            )
        except P4ClientError as ex:
            if strict or not ex.is_not_found_only:
                if warn_only:
                    LOGGER.warning(
                        "WriteRequest with `warn_only=True` failed",
                        exc_info=True,
                    )
                else:
                    raise

            assert (not strict and ex.is_not_found_only) or warn_only

    async def _fetch_capabilities(self):
        "Check the P4Runtime protocol version supported by the other end."
        assert self._p4client is not None

        try:
            reply = await self._p4client.request(p4r.CapabilitiesRequest())
            self._api_version = ApiVersion.parse(reply.p4runtime_api_version)

        except P4ClientError as ex:
            if ex.code != GRPCStatusCode.UNIMPLEMENTED:
                raise
            LOGGER.warning("CapabilitiesRequest is not implemented")

    async def _start_gnmi(self):
        "Start the associated gNMI client."
        assert self._gnmi_client is None
        assert self._p4client is not None

        self._gnmi_client = GNMIClient(self._address, self._options.channel_credentials)
        await self._gnmi_client.open(channel=self._p4client.channel)

        try:
            await self._ports.subscribe(self._gnmi_client)
            if self._ports:
                self.create_task(self._ports.listen(), background=True, name="_ports")

        except GNMIClientError as ex:
            if ex.code != GRPCStatusCode.UNIMPLEMENTED:
                raise
            LOGGER.warning("gNMI is not implemented")
            await self._gnmi_client.close()
            self._gnmi_client = None

    async def _stop_gnmi(self):
        "Stop the associated gNMI client."
        if self._gnmi_client is not None:
            self._ports.close()
            await self._gnmi_client.close()
            self._gnmi_client = None

    def __repr__(self) -> str:
        "Return string representation of switch."
        return f"Switch(name={self._name!r}, address={self._address!r})"


class SwitchEvent(str, enum.Enum):
    "Events for Switch class."

    CONTROLLER_ENTER = "controller_enter"  # (switch)
    CONTROLLER_LEAVE = "controller_leave"  # (switch)
    SWITCH_START = "switch_start"  # (switch)
    SWITCH_STOP = "switch_stop"  # (switch)
    CHANNEL_UP = "channel_up"  # (switch)
    CHANNEL_DOWN = "channel_down"  # (switch)
    CHANNEL_READY = "channel_ready"  # (switch)
    BECOME_PRIMARY = "become_primary"  # (switch)
    BECOME_BACKUP = "become_backup"  # (switch)
    PORT_UP = "port_up"  # (switch, port)
    PORT_DOWN = "port_down"  # (switch, port)
    STREAM_ERROR = "stream_error"  # (switch, p4r.StreamMessageResponse)


class SwitchEmitter(pyee.EventEmitter):
    "EventEmitter subclass for emitting Switch events."

    def __init__(self, switch: Switch):
        super().__init__()
        self._switch = switch

    def _emit_run(self, f: Any, args: Any, kwargs: Any):
        try:
            coro: Any = f(*args, **kwargs)
        except Exception as exc:  # pylint: disable=broad-except
            LOGGER.error("Error in SwitchEmitter._emit_run", exc_info=exc)
            # TODO: Have synchronous callback failures affect Switch?
        else:
            if asyncio.iscoroutine(coro):
                self._switch.create_task(coro)

    def event_future(self, event: SwitchEvent) -> asyncio.Future[Any]:
        "Future to wait for a specific event."
        ready = asyncio.get_running_loop().create_future()
        self.once(event, lambda _: ready.set_result(None))  # type: ignore
        return ready


class SwitchFailFastError(BaseException):
    "This exception is raised when a Switch with `fail_fast=True` fails."


class SwitchTasks:
    "Manage a set of related tasks."

    _tasks: set[asyncio.Task[Any]]
    _task_count: CountdownFuture
    _fail_fast: bool
    _fail_count: int

    def __init__(self, fail_fast: bool):
        self._tasks = set()
        self._task_count = CountdownFuture()
        self._fail_fast = fail_fast
        self._fail_count = 0

    def create_task(
        self,
        coro: Coroutine[Any, Any, _T],
        *,
        switch: Switch,
        background: bool,
        name: str | None = None,
    ) -> asyncio.Task[_T]:
        "Create a new task managed by switch."
        if name is None:
            name = coro.__qualname__

        bg_char = "&" if background else ""
        task_name = f"fy:{switch.name}|{name}{bg_char}"

        task = asyncio.create_task(coro, name=task_name)
        self._tasks.add(task)
        self._task_count.increment()
        task.add_done_callback(self._task_done)

        return task

    def _task_done(self, done: asyncio.Task[Any]):
        "Called when a task finishes."
        self._tasks.remove(done)
        self._task_count.decrement()

        if not done.cancelled():
            ex = done.exception()
            if ex is not None:
                self.cancel_all()

                # If the exception is GRPCStatusCode.UNAVAILABLE, don't report
                # an error.
                if getattr(ex, "code", None) == GRPCStatusCode.UNAVAILABLE:
                    LOGGER.debug("Switch task %r failed: UNAVAILABLE", done.get_name())
                else:
                    self._fail_count += 1
                    LOGGER.critical(
                        "Switch task %r failed", done.get_name(), exc_info=ex
                    )

    def cancel_all(self) -> None:
        "Cancel all tasks."
        for task in self._tasks:
            if not task.done():
                task.cancel()

    def cancel_primary(self) -> None:
        "Cancel all non-background tasks."
        for task in self._tasks:
            if not task.done() and not task.get_name().endswith("&"):
                task.cancel()

    async def wait(self) -> None:
        "Wait for all tasks to finish."
        try:
            await self._task_count.wait(self.cancel_all)
        finally:
            if self._fail_fast and self._fail_count > 0:
                raise SwitchFailFastError()


@asynccontextmanager
async def _throttle_failure():
    "Used to throttle retries, if the switch's lifecycle fails too quickly."
    start_time = time.monotonic()
    yield
    end_time = time.monotonic()

    if end_time - start_time < _FAIL_MIN_TIME_SECS:
        # Handle case where we failed too quickly.
        LOGGER.warning(
            "Switch failed in less than %g seconds; waiting %g seconds to try again",
            _FAIL_MIN_TIME_SECS,
            _FAIL_THROTTLE_SECS,
        )
        await asyncio.sleep(_FAIL_THROTTLE_SECS)

    else:
        # Otherwise, always yield to other tasks here.
        await asyncio.sleep(0)



"Implement P4TypeFactory helper."

import finsy.p4schema as sch
from finsy.p4values import encode_enum
from finsy.proto import p4t


class P4TypeFactory:
    """Factory class to dynamically construct P4 types during tests.

    Examples:

        factory = P4TypeFactory()
        bool_t = factory.bool_type()
        uint8_t = factory.bits_type(8)
        my_struct = factory.struct_type("my_struct", x=bool_t, y=uint8_t)
        my_enum = factory.serializable_enum("my_enum", 32, X=1, Y=2)
        data = my_struct.encode_data({"x": True, "y": 100})
        print(data)
    """

    _type_info: sch.P4TypeInfo

    def __init__(self):
        self._type_info = sch.P4TypeInfo(p4t.P4TypeInfo())

    def bool_type(self) -> sch.P4BoolType:
        "Construct a `bool` type."
        return sch.P4BoolType(p4t.P4BoolType())

    def bits_type(
        self,
        bitwidth: int,
        *,
        signed: bool = False,
        varbit: bool = False,
    ) -> sch.P4BitsType:
        "Construct a `bits` type, which may be signed or varbit."
        if varbit:
            assert not signed
            return sch.P4BitsType(
                p4t.P4BitstringLikeTypeSpec(
                    varbit=p4t.P4VarbitTypeSpec(max_bitwidth=bitwidth)
                )
            )

        if signed:
            assert not varbit
            return sch.P4BitsType(
                p4t.P4BitstringLikeTypeSpec(int=p4t.P4IntTypeSpec(bitwidth=bitwidth))
            )

        return sch.P4BitsType(
            p4t.P4BitstringLikeTypeSpec(bit=p4t.P4BitTypeSpec(bitwidth=bitwidth))
        )

    def tuple_type(self, *members: sch.P4Type) -> sch.P4TupleType:
        "Construct a `tuple` type."
        return sch.P4TupleType(
            p4t.P4TupleTypeSpec(members=(member.data_type_spec for member in members)),
            type_info=self._type_info,
        )

    def struct_type(self, __name: str, **members: sch.P4Type) -> sch.P4StructType:
        "Construct a `struct` type."
        result = sch.P4StructType(
            __name,
            p4t.P4StructTypeSpec(
                members=[
                    p4t.P4StructTypeSpec.Member(
                        name=key, type_spec=value.data_type_spec
                    )
                    for key, value in members.items()
                ]
            ),
        )
        self._type_info.structs[__name] = result
        result._finish_init(self._type_info)  # pyright: ignore[reportPrivateUsage]
        return result

    def serializable_enum_type(
        self,
        __name: str,
        __bitwidth: int,
        **members: int,
    ) -> sch.P4SerializableEnumType:
        "Construct a `serializable_enum` type."
        result = sch.P4SerializableEnumType(
            __name,
            p4t.P4SerializableEnumTypeSpec(
                underlying_type=p4t.P4BitTypeSpec(bitwidth=__bitwidth),
                members=[
                    p4t.P4SerializableEnumTypeSpec.Member(
                        name=name,
                        value=encode_enum(value, __bitwidth),
                    )
                    for name, value in members.items()
                ],
            ),
        )
        self._type_info.serializable_enums[__name] = result
        return result



"Implements string utility functions."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def minimum_edit_distance(str1: str, str2: str) -> int:
    """Compute the minimum edit distance between two strings.

    Computes the Levenshtein distance using the iterative Wagner-Fischer
    algorithm.

    Reference:  https://en.wikipedia.org/wiki/Levenshtein_distance
      See "Iterative with two matrix rows"
      Replace `v1` accesses with `prev` to eliminate second row `v1`.
    """
    if str1 == str2:
        return 0

    len1 = len(str1)
    len2 = len(str2)
    row = list(range(len2 + 1))

    for i in range(len1):
        prev = i + 1
        for j in range(len2):
            min_cost = row[j]
            if str1[i] != str2[j]:
                min_cost += 1

                delete = row[j + 1] + 1
                if delete < min_cost:
                    min_cost = delete

                insert = prev + 1
                if insert < min_cost:
                    min_cost = insert

            row[j] = prev
            prev = min_cost

        row[len2] = prev

    return row[len2]



"Implements the P4Client class."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from dataclasses import dataclass
from typing import Any, AsyncIterator, Callable, Sequence, TypeAlias, cast, overload

import grpc  # pyright: ignore[reportMissingTypeStubs]
from typing_extensions import Self

from finsy import pbutil
from finsy.grpcutil import (
    GRPC_EOF,
    GRPCCredentialsTLS,
    GRPCOptions,
    GRPCStatusCode,
    grpc_channel,
)
from finsy.log import LOGGER
from finsy.p4schema import P4Schema
from finsy.proto import p4r, p4r_grpc, rpc_code, rpc_status

_DEFAULT_RPC_TIMEOUT = 10.0

# GRPCStatusCode.INVALID_ARGUMENT
#   simple_switch_grpc: 'Election id already exists'
#   stratum_bmv2: 'Election ID is already used by another connection with the same role.'

_ELECTION_ID_EXISTS = re.compile(
    r"election id .*\b(?:used|exists)\b",
    re.IGNORECASE,
)

# GRPCStatusCode.FAILED_PRECONDITION
#   simple_switch_grpc: 'No forwarding pipeline config set for this device.'
#   stratum_bmv2: 'No valid forwarding pipeline config has been pushed for any node so far.'

_NO_PIPELINE_CONFIG = re.compile(
    r"no .*\bforwarding pipeline config",
    re.IGNORECASE,
)


@dataclass
class P4Error:
    "P4Runtime Error message used to report a single P4-entity error."

    canonical_code: GRPCStatusCode
    message: str
    space: str
    code: int
    subvalue: pbutil.PBMessage | None = None


@dataclass
class P4RpcStatus:
    "Implements rpc.status."

    code: GRPCStatusCode
    message: str
    details: dict[int, P4Error]

    @staticmethod
    def from_rpc_error(error: grpc.RpcError) -> "P4RpcStatus":
        "Construct status from RpcError."
        assert isinstance(error, grpc.aio.AioRpcError)

        for key, value in error.trailing_metadata():
            if key == "grpc-status-details-bin":
                assert isinstance(value, bytes)
                return P4RpcStatus.from_bytes(value)

        return P4RpcStatus(
            GRPCStatusCode.from_status_code(error.code()),
            error.details() or "",
            {},
        )

    @staticmethod
    def from_bytes(data: bytes) -> "P4RpcStatus":
        "Construct status from protobuf bytes."
        status = rpc_status.Status()
        status.ParseFromString(data)
        return P4RpcStatus.from_status(status)

    @staticmethod
    def from_status(status: rpc_status.Status) -> "P4RpcStatus":
        "Construct status from RPC status."
        return P4RpcStatus(
            GRPCStatusCode(status.code),
            status.message,
            P4RpcStatus._parse_error(status.details),
        )

    @staticmethod
    def _parse_error(details: Sequence[pbutil.PBAny]) -> dict[int, P4Error]:
        result: dict[int, P4Error] = {}

        for i, detail in enumerate(details):
            err = pbutil.from_any(detail, p4r.Error)
            if err.canonical_code != rpc_code.OK:
                result[i] = P4Error(
                    GRPCStatusCode(err.canonical_code),
                    err.message,
                    err.space,
                    err.code,
                )
        return result


class P4ClientError(Exception):
    "Wrap `grpc.RpcError`."

    _operation: str
    _status: P4RpcStatus
    _outer_code: GRPCStatusCode
    _outer_message: str
    _schema: P4Schema | None = None  # for annotating sub-value details

    def __init__(
        self,
        error: grpc.RpcError,
        operation: str,
        *,
        msg: pbutil.PBMessage | None = None,
        schema: P4Schema | None = None,
    ):
        super().__init__()
        assert isinstance(error, grpc.aio.AioRpcError)

        self._operation = operation
        self._status = P4RpcStatus.from_rpc_error(error)
        self._outer_code = GRPCStatusCode.from_status_code(error.code())
        self._outer_message = error.details() or ""

        if msg is not None and self.details:
            self._attach_details(msg)
            self._schema = schema

        LOGGER.debug("%s failed: %s", operation, self)

    @property
    def code(self) -> GRPCStatusCode:
        "GRPC status code."
        return self._status.code

    @property
    def message(self) -> str:
        "GRPC status message."
        return self._status.message

    @property
    def details(self) -> dict[int, P4Error]:
        "Optional details about P4Runtime Write updates that failed."
        return self._status.details

    @property
    def is_not_found_only(self) -> bool:
        """Return True if the only sub-errors are NOT_FOUND."""
        if self.code != GRPCStatusCode.UNKNOWN:
            return False

        for err in self.details.values():
            if err.canonical_code != GRPCStatusCode.NOT_FOUND:
                return False
        return True

    @property
    def is_election_id_used(self) -> bool:
        """Return true if error is that election ID is in use."""
        return (
            self.code == GRPCStatusCode.INVALID_ARGUMENT
            and _ELECTION_ID_EXISTS.search(self.message) is not None
        )

    @property
    def is_pipeline_missing(self) -> bool:
        "Return true if error is that no pipeline config is set."
        return (
            self.code == GRPCStatusCode.FAILED_PRECONDITION
            and _NO_PIPELINE_CONFIG.search(self.message) is not None
        )

    def _attach_details(self, msg: pbutil.PBMessage):
        "Attach the subvalue(s) from the message that caused the error."
        if isinstance(msg, p4r.WriteRequest):
            for key, value in self.details.items():
                value.subvalue = msg.updates[key]

    def __str__(self) -> str:
        "Return string representation of P4ClientError object."
        if self.details:

            def _show(value: P4Error):
                s = repr(value)
                if self._schema:
                    s = pbutil.log_annotate(s, self._schema)
                s = s.replace("\n}\n)", "\n})")  # tidy multiline repr
                return s.replace("\n", "\n" + " " * 6)  # indent 6 spaces

            items = [""] + [
                f"  [details.{key}] {_show(val)}" for key, val in self.details.items()
            ]
            details = "\n".join(items)
        else:
            details = ""

        if self.code == self._outer_code and self.message == self._outer_message:
            return (
                f"operation={self._operation} code={self.code!r} "
                f"message={self.message!r} {details}"
            )
        return (
            f"code={self.code!r} message={self.message!r} "
            f"details={self.details!r} operation={self._operation} "
            f"_outer_message={self._outer_message!r} _outer_code={self._outer_code!r}"
        )


_P4StreamTypeAlias: TypeAlias = grpc.aio.StreamStreamCall[
    p4r.StreamMessageRequest, p4r.StreamMessageResponse
]


class P4Client:
    "Implements a P4Runtime client."

    _address: str
    _credentials: GRPCCredentialsTLS | None
    _wait_for_ready: bool
    _channel: grpc.aio.Channel | None = None
    _stub: p4r_grpc.P4RuntimeStub | None = None
    _stream: _P4StreamTypeAlias | None = None
    _complete_request: Callable[[pbutil.PBMessage], None] | None = None

    _schema: P4Schema | None = None
    "Annotate log messages using this optional P4Info schema."

    def __init__(
        self,
        address: str,
        credentials: GRPCCredentialsTLS | None = None,
        *,
        wait_for_ready: bool = True,
    ) -> None:
        self._address = address
        self._credentials = credentials
        self._wait_for_ready = wait_for_ready

    @property
    def channel(self) -> grpc.aio.Channel | None:
        "Return the GRPC channel object, or None if the channel is not open."
        return self._channel

    async def __aenter__(self) -> Self:
        await self.open()
        return self

    async def __aexit__(self, *args: Any) -> bool | None:
        await self.close()

    async def open(
        self,
        *,
        schema: P4Schema | None = None,
        complete_request: Callable[[pbutil.PBMessage], None] | None = None,
    ) -> None:
        """Open the client channel.

        Note: This method is `async` for forward-compatible reasons.
        """
        assert self._channel is None
        assert self._stub is None

        # Increase max_metadata_size from 8 KB to 64 KB.
        options = GRPCOptions(
            max_metadata_size=64 * 1024,  # 64 kilobytes
            max_reconnect_backoff_ms=15000,  # 15.0 seconds
        )

        self._channel = grpc_channel(
            self._address,
            credentials=self._credentials,
            options=options,
            client_type="P4Client",
        )

        self._stub = p4r_grpc.P4RuntimeStub(self._channel)
        self._schema = schema
        self._complete_request = complete_request

    async def close(self) -> None:
        "Close the client channel."
        if self._channel is not None:
            LOGGER.debug("P4Client: close channel %r", self._address)

            if self._stream is not None:
                self._stream.cancel()
                self._stream = None

            await self._channel.close()
            self._channel = None
            self._stub = None
            self._schema = None
            self._complete_request = None

    async def send(self, msg: p4r.StreamMessageRequest) -> None:
        """Send a message to the stream."""
        assert self._stub is not None

        if not self._stream or self._stream.done():
            self._stream = cast(
                _P4StreamTypeAlias,
                self._stub.StreamChannel(wait_for_ready=self._wait_for_ready),  # type: ignore
            )

        self._log_msg(msg)

        try:
            await self._stream.write(msg)
        except grpc.RpcError as ex:
            raise P4ClientError(ex, "send") from None

    async def receive(self) -> p4r.StreamMessageResponse:
        """Read a message from the stream."""
        assert self._stream is not None

        try:
            msg = cast(
                p4r.StreamMessageResponse,
                await self._stream.read(),  # type: ignore
            )
            if msg is GRPC_EOF:
                # Treat EOF as a protocol violation.
                raise RuntimeError("P4Client.receive got EOF!")

        except grpc.RpcError as ex:
            raise P4ClientError(ex, "receive") from None

        self._log_msg(msg)
        return msg

    @overload
    async def request(
        self, msg: p4r.WriteRequest
    ) -> p4r.WriteResponse: ...  # pragma: no cover

    @overload
    async def request(
        self, msg: p4r.GetForwardingPipelineConfigRequest
    ) -> p4r.GetForwardingPipelineConfigResponse: ...  # pragma: no cover

    @overload
    async def request(
        self, msg: p4r.SetForwardingPipelineConfigRequest
    ) -> p4r.SetForwardingPipelineConfigResponse: ...  # pragma: no cover

    @overload
    async def request(
        self, msg: p4r.CapabilitiesRequest
    ) -> p4r.CapabilitiesResponse: ...  # pragma: no cover

    async def request(self, msg: pbutil.PBMessage) -> pbutil.PBMessage:
        "Send a unary-unary P4Runtime request and wait for the response."
        assert self._stub is not None

        if self._complete_request:
            self._complete_request(msg)

        msg_type = type(msg).__name__
        assert msg_type.endswith("Request")
        rpc_method = getattr(self._stub, msg_type[:-7])

        self._log_msg(msg)
        try:
            reply = await rpc_method(
                msg,
                timeout=_DEFAULT_RPC_TIMEOUT,
            )
        except grpc.RpcError as ex:
            raise P4ClientError(ex, msg_type, msg=msg, schema=self._schema) from None

        self._log_msg(reply)
        return reply

    async def request_iter(
        self, msg: p4r.ReadRequest
    ) -> AsyncIterator[p4r.ReadResponse]:
        "Send a unary-stream P4Runtime read request and wait for the responses."
        assert self._stub is not None

        if self._complete_request:
            self._complete_request(msg)

        msg_type = type(msg).__name__
        assert msg_type.endswith("Request")
        rpc_method = getattr(self._stub, msg_type[:-7])

        self._log_msg(msg)
        try:
            async for reply in rpc_method(
                msg,
                timeout=_DEFAULT_RPC_TIMEOUT,
            ):
                self._log_msg(reply)
                yield reply
        except grpc.RpcError as ex:
            raise P4ClientError(ex, msg_type) from None

    def _log_msg(self, msg: pbutil.PBMessage) -> None:
        "Log a P4Runtime request or response."
        pbutil.log_msg(self._channel, msg, self._schema)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gnmi1/gnmi.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2
from . import gnmi_ext_pb2 as gnmi1_dot_gnmi__ext__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10gnmi1/gnmi.proto\x12\x04gnmi\x1a\x19google/protobuf/any.proto\x1a google/protobuf/descriptor.proto\x1a\x14gnmi1/gnmi_ext.proto\"\x94\x01\n\x0cNotification\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\x1a\n\x06prefix\x18\x02 \x01(\x0b\x32\n.gnmi.Path\x12\x1c\n\x06update\x18\x04 \x03(\x0b\x32\x0c.gnmi.Update\x12\x1a\n\x06\x64\x65lete\x18\x05 \x03(\x0b\x32\n.gnmi.Path\x12\x0e\n\x06\x61tomic\x18\x06 \x01(\x08J\x04\x08\x03\x10\x04R\x05\x61lias\"u\n\x06Update\x12\x18\n\x04path\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12\x1e\n\x05value\x18\x02 \x01(\x0b\x32\x0b.gnmi.ValueB\x02\x18\x01\x12\x1d\n\x03val\x18\x03 \x01(\x0b\x32\x10.gnmi.TypedValue\x12\x12\n\nduplicates\x18\x04 \x01(\r\"\x83\x03\n\nTypedValue\x12\x14\n\nstring_val\x18\x01 \x01(\tH\x00\x12\x11\n\x07int_val\x18\x02 \x01(\x03H\x00\x12\x12\n\x08uint_val\x18\x03 \x01(\x04H\x00\x12\x12\n\x08\x62ool_val\x18\x04 \x01(\x08H\x00\x12\x13\n\tbytes_val\x18\x05 \x01(\x0cH\x00\x12\x17\n\tfloat_val\x18\x06 \x01(\x02\x42\x02\x18\x01H\x00\x12\x14\n\ndouble_val\x18\x0e \x01(\x01H\x00\x12*\n\x0b\x64\x65\x63imal_val\x18\x07 \x01(\x0b\x32\x0f.gnmi.Decimal64B\x02\x18\x01H\x00\x12)\n\x0cleaflist_val\x18\x08 \x01(\x0b\x32\x11.gnmi.ScalarArrayH\x00\x12\'\n\x07\x61ny_val\x18\t \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x12\x12\n\x08json_val\x18\n \x01(\x0cH\x00\x12\x17\n\rjson_ietf_val\x18\x0b \x01(\x0cH\x00\x12\x13\n\tascii_val\x18\x0c \x01(\tH\x00\x12\x15\n\x0bproto_bytes\x18\r \x01(\x0cH\x00\x42\x07\n\x05value\"Y\n\x04Path\x12\x13\n\x07\x65lement\x18\x01 \x03(\tB\x02\x18\x01\x12\x0e\n\x06origin\x18\x02 \x01(\t\x12\x1c\n\x04\x65lem\x18\x03 \x03(\x0b\x32\x0e.gnmi.PathElem\x12\x0e\n\x06target\x18\x04 \x01(\t\"j\n\x08PathElem\x12\x0c\n\x04name\x18\x01 \x01(\t\x12$\n\x03key\x18\x02 \x03(\x0b\x32\x17.gnmi.PathElem.KeyEntry\x1a*\n\x08KeyEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"8\n\x05Value\x12\r\n\x05value\x18\x01 \x01(\x0c\x12\x1c\n\x04type\x18\x02 \x01(\x0e\x32\x0e.gnmi.Encoding:\x02\x18\x01\"N\n\x05\x45rror\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\"\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\x14.google.protobuf.Any:\x02\x18\x01\"2\n\tDecimal64\x12\x0e\n\x06\x64igits\x18\x01 \x01(\x03\x12\x11\n\tprecision\x18\x02 \x01(\r:\x02\x18\x01\"0\n\x0bScalarArray\x12!\n\x07\x65lement\x18\x01 \x03(\x0b\x32\x10.gnmi.TypedValue\"\x9d\x01\n\x10SubscribeRequest\x12+\n\tsubscribe\x18\x01 \x01(\x0b\x32\x16.gnmi.SubscriptionListH\x00\x12\x1a\n\x04poll\x18\x03 \x01(\x0b\x32\n.gnmi.PollH\x00\x12&\n\textension\x18\x05 \x03(\x0b\x32\x13.gnmi_ext.ExtensionB\t\n\x07requestJ\x04\x08\x04\x10\x05R\x07\x61liases\"\x06\n\x04Poll\"\xa8\x01\n\x11SubscribeResponse\x12$\n\x06update\x18\x01 \x01(\x0b\x32\x12.gnmi.NotificationH\x00\x12\x17\n\rsync_response\x18\x03 \x01(\x08H\x00\x12 \n\x05\x65rror\x18\x04 \x01(\x0b\x32\x0b.gnmi.ErrorB\x02\x18\x01H\x00\x12&\n\textension\x18\x05 \x03(\x0b\x32\x13.gnmi_ext.ExtensionB\n\n\x08response\"\xd5\x02\n\x10SubscriptionList\x12\x1a\n\x06prefix\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12(\n\x0csubscription\x18\x02 \x03(\x0b\x32\x12.gnmi.Subscription\x12\x1d\n\x03qos\x18\x04 \x01(\x0b\x32\x10.gnmi.QOSMarking\x12)\n\x04mode\x18\x05 \x01(\x0e\x32\x1b.gnmi.SubscriptionList.Mode\x12\x19\n\x11\x61llow_aggregation\x18\x06 \x01(\x08\x12#\n\nuse_models\x18\x07 \x03(\x0b\x32\x0f.gnmi.ModelData\x12 \n\x08\x65ncoding\x18\x08 \x01(\x0e\x32\x0e.gnmi.Encoding\x12\x14\n\x0cupdates_only\x18\t \x01(\x08\"&\n\x04Mode\x12\n\n\x06STREAM\x10\x00\x12\x08\n\x04ONCE\x10\x01\x12\x08\n\x04POLL\x10\x02J\x04\x08\x03\x10\x04R\x0buse_aliases\"\x9f\x01\n\x0cSubscription\x12\x18\n\x04path\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12$\n\x04mode\x18\x02 \x01(\x0e\x32\x16.gnmi.SubscriptionMode\x12\x17\n\x0fsample_interval\x18\x03 \x01(\x04\x12\x1a\n\x12suppress_redundant\x18\x04 \x01(\x08\x12\x1a\n\x12heartbeat_interval\x18\x05 \x01(\x04\"\x1d\n\nQOSMarking\x12\x0f\n\x07marking\x18\x01 \x01(\r\"\xce\x01\n\nSetRequest\x12\x1a\n\x06prefix\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12\x1a\n\x06\x64\x65lete\x18\x02 \x03(\x0b\x32\n.gnmi.Path\x12\x1d\n\x07replace\x18\x03 \x03(\x0b\x32\x0c.gnmi.Update\x12\x1c\n\x06update\x18\x04 \x03(\x0b\x32\x0c.gnmi.Update\x12#\n\runion_replace\x18\x06 \x03(\x0b\x32\x0c.gnmi.Update\x12&\n\textension\x18\x05 \x03(\x0b\x32\x13.gnmi_ext.Extension\"\xac\x01\n\x0bSetResponse\x12\x1a\n\x06prefix\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12$\n\x08response\x18\x02 \x03(\x0b\x32\x12.gnmi.UpdateResult\x12 \n\x07message\x18\x03 \x01(\x0b\x32\x0b.gnmi.ErrorB\x02\x18\x01\x12\x11\n\ttimestamp\x18\x04 \x01(\x03\x12&\n\textension\x18\x05 \x03(\x0b\x32\x13.gnmi_ext.Extension\"\xdd\x01\n\x0cUpdateResult\x12\x15\n\ttimestamp\x18\x01 \x01(\x03\x42\x02\x18\x01\x12\x18\n\x04path\x18\x02 \x01(\x0b\x32\n.gnmi.Path\x12 \n\x07message\x18\x03 \x01(\x0b\x32\x0b.gnmi.ErrorB\x02\x18\x01\x12(\n\x02op\x18\x04 \x01(\x0e\x32\x1c.gnmi.UpdateResult.Operation\"P\n\tOperation\x12\x0b\n\x07INVALID\x10\x00\x12\n\n\x06\x44\x45LETE\x10\x01\x12\x0b\n\x07REPLACE\x10\x02\x12\n\n\x06UPDATE\x10\x03\x12\x11\n\rUNION_REPLACE\x10\x04\"\x97\x02\n\nGetRequest\x12\x1a\n\x06prefix\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12\x18\n\x04path\x18\x02 \x03(\x0b\x32\n.gnmi.Path\x12\'\n\x04type\x18\x03 \x01(\x0e\x32\x19.gnmi.GetRequest.DataType\x12 \n\x08\x65ncoding\x18\x05 \x01(\x0e\x32\x0e.gnmi.Encoding\x12#\n\nuse_models\x18\x06 \x03(\x0b\x32\x0f.gnmi.ModelData\x12&\n\textension\x18\x07 \x03(\x0b\x32\x13.gnmi_ext.Extension\";\n\x08\x44\x61taType\x12\x07\n\x03\x41LL\x10\x00\x12\n\n\x06\x43ONFIG\x10\x01\x12\t\n\x05STATE\x10\x02\x12\x0f\n\x0bOPERATIONAL\x10\x03\"\x7f\n\x0bGetResponse\x12(\n\x0cnotification\x18\x01 \x03(\x0b\x32\x12.gnmi.Notification\x12\x1e\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x0b.gnmi.ErrorB\x02\x18\x01\x12&\n\textension\x18\x03 \x03(\x0b\x32\x13.gnmi_ext.Extension\";\n\x11\x43\x61pabilityRequest\x12&\n\textension\x18\x01 \x03(\x0b\x32\x13.gnmi_ext.Extension\"\xaa\x01\n\x12\x43\x61pabilityResponse\x12)\n\x10supported_models\x18\x01 \x03(\x0b\x32\x0f.gnmi.ModelData\x12+\n\x13supported_encodings\x18\x02 \x03(\x0e\x32\x0e.gnmi.Encoding\x12\x14\n\x0cgNMI_version\x18\x03 \x01(\t\x12&\n\textension\x18\x04 \x03(\x0b\x32\x13.gnmi_ext.Extension\"@\n\tModelData\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0corganization\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t*D\n\x08\x45ncoding\x12\x08\n\x04JSON\x10\x00\x12\t\n\x05\x42YTES\x10\x01\x12\t\n\x05PROTO\x10\x02\x12\t\n\x05\x41SCII\x10\x03\x12\r\n\tJSON_IETF\x10\x04*A\n\x10SubscriptionMode\x12\x12\n\x0eTARGET_DEFINED\x10\x00\x12\r\n\tON_CHANGE\x10\x01\x12\n\n\x06SAMPLE\x10\x02\x32\xe3\x01\n\x04gNMI\x12\x41\n\x0c\x43\x61pabilities\x12\x17.gnmi.CapabilityRequest\x1a\x18.gnmi.CapabilityResponse\x12*\n\x03Get\x12\x10.gnmi.GetRequest\x1a\x11.gnmi.GetResponse\x12*\n\x03Set\x12\x10.gnmi.SetRequest\x1a\x11.gnmi.SetResponse\x12@\n\tSubscribe\x12\x16.gnmi.SubscribeRequest\x1a\x17.gnmi.SubscribeResponse(\x01\x30\x01:3\n\x0cgnmi_service\x12\x1c.google.protobuf.FileOptions\x18\xe9\x07 \x01(\tBT\n\x15\x63om.github.gnmi.protoB\tGnmiProtoP\x01Z%github.com/openconfig/gnmi/proto/gnmi\xca>\x06\x30.10.0b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'gnmi1.gnmi_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\025com.github.gnmi.protoB\tGnmiProtoP\001Z%github.com/openconfig/gnmi/proto/gnmi\312>\0060.10.0'
  _globals['_UPDATE'].fields_by_name['value']._loaded_options = None
  _globals['_UPDATE'].fields_by_name['value']._serialized_options = b'\030\001'
  _globals['_TYPEDVALUE'].fields_by_name['float_val']._loaded_options = None
  _globals['_TYPEDVALUE'].fields_by_name['float_val']._serialized_options = b'\030\001'
  _globals['_TYPEDVALUE'].fields_by_name['decimal_val']._loaded_options = None
  _globals['_TYPEDVALUE'].fields_by_name['decimal_val']._serialized_options = b'\030\001'
  _globals['_PATH'].fields_by_name['element']._loaded_options = None
  _globals['_PATH'].fields_by_name['element']._serialized_options = b'\030\001'
  _globals['_PATHELEM_KEYENTRY']._loaded_options = None
  _globals['_PATHELEM_KEYENTRY']._serialized_options = b'8\001'
  _globals['_VALUE']._loaded_options = None
  _globals['_VALUE']._serialized_options = b'\030\001'
  _globals['_ERROR']._loaded_options = None
  _globals['_ERROR']._serialized_options = b'\030\001'
  _globals['_DECIMAL64']._loaded_options = None
  _globals['_DECIMAL64']._serialized_options = b'\030\001'
  _globals['_SUBSCRIBERESPONSE'].fields_by_name['error']._loaded_options = None
  _globals['_SUBSCRIBERESPONSE'].fields_by_name['error']._serialized_options = b'\030\001'
  _globals['_SETRESPONSE'].fields_by_name['message']._loaded_options = None
  _globals['_SETRESPONSE'].fields_by_name['message']._serialized_options = b'\030\001'
  _globals['_UPDATERESULT'].fields_by_name['timestamp']._loaded_options = None
  _globals['_UPDATERESULT'].fields_by_name['timestamp']._serialized_options = b'\030\001'
  _globals['_UPDATERESULT'].fields_by_name['message']._loaded_options = None
  _globals['_UPDATERESULT'].fields_by_name['message']._serialized_options = b'\030\001'
  _globals['_GETRESPONSE'].fields_by_name['error']._loaded_options = None
  _globals['_GETRESPONSE'].fields_by_name['error']._serialized_options = b'\030\001'
  _globals['_ENCODING']._serialized_start=3403
  _globals['_ENCODING']._serialized_end=3471
  _globals['_SUBSCRIPTIONMODE']._serialized_start=3473
  _globals['_SUBSCRIPTIONMODE']._serialized_end=3538
  _globals['_NOTIFICATION']._serialized_start=110
  _globals['_NOTIFICATION']._serialized_end=258
  _globals['_UPDATE']._serialized_start=260
  _globals['_UPDATE']._serialized_end=377
  _globals['_TYPEDVALUE']._serialized_start=380
  _globals['_TYPEDVALUE']._serialized_end=767
  _globals['_PATH']._serialized_start=769
  _globals['_PATH']._serialized_end=858
  _globals['_PATHELEM']._serialized_start=860
  _globals['_PATHELEM']._serialized_end=966
  _globals['_PATHELEM_KEYENTRY']._serialized_start=924
  _globals['_PATHELEM_KEYENTRY']._serialized_end=966
  _globals['_VALUE']._serialized_start=968
  _globals['_VALUE']._serialized_end=1024
  _globals['_ERROR']._serialized_start=1026
  _globals['_ERROR']._serialized_end=1104
  _globals['_DECIMAL64']._serialized_start=1106
  _globals['_DECIMAL64']._serialized_end=1156
  _globals['_SCALARARRAY']._serialized_start=1158
  _globals['_SCALARARRAY']._serialized_end=1206
  _globals['_SUBSCRIBEREQUEST']._serialized_start=1209
  _globals['_SUBSCRIBEREQUEST']._serialized_end=1366
  _globals['_POLL']._serialized_start=1368
  _globals['_POLL']._serialized_end=1374
  _globals['_SUBSCRIBERESPONSE']._serialized_start=1377
  _globals['_SUBSCRIBERESPONSE']._serialized_end=1545
  _globals['_SUBSCRIPTIONLIST']._serialized_start=1548
  _globals['_SUBSCRIPTIONLIST']._serialized_end=1889
  _globals['_SUBSCRIPTIONLIST_MODE']._serialized_start=1832
  _globals['_SUBSCRIPTIONLIST_MODE']._serialized_end=1870
  _globals['_SUBSCRIPTION']._serialized_start=1892
  _globals['_SUBSCRIPTION']._serialized_end=2051
  _globals['_QOSMARKING']._serialized_start=2053
  _globals['_QOSMARKING']._serialized_end=2082
  _globals['_SETREQUEST']._serialized_start=2085
  _globals['_SETREQUEST']._serialized_end=2291
  _globals['_SETRESPONSE']._serialized_start=2294
  _globals['_SETRESPONSE']._serialized_end=2466
  _globals['_UPDATERESULT']._serialized_start=2469
  _globals['_UPDATERESULT']._serialized_end=2690
  _globals['_UPDATERESULT_OPERATION']._serialized_start=2610
  _globals['_UPDATERESULT_OPERATION']._serialized_end=2690
  _globals['_GETREQUEST']._serialized_start=2693
  _globals['_GETREQUEST']._serialized_end=2972
  _globals['_GETREQUEST_DATATYPE']._serialized_start=2913
  _globals['_GETREQUEST_DATATYPE']._serialized_end=2972
  _globals['_GETRESPONSE']._serialized_start=2974
  _globals['_GETRESPONSE']._serialized_end=3101
  _globals['_CAPABILITYREQUEST']._serialized_start=3103
  _globals['_CAPABILITYREQUEST']._serialized_end=3162
  _globals['_CAPABILITYRESPONSE']._serialized_start=3165
  _globals['_CAPABILITYRESPONSE']._serialized_end=3335
  _globals['_MODELDATA']._serialized_start=3337
  _globals['_MODELDATA']._serialized_end=3401
  _globals['_GNMI']._serialized_start=3541
  _globals['_GNMI']._serialized_end=3768
# @@protoc_insertion_point(module_scope)



"Implements the run() function."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import signal
import sys
from typing import Any, Coroutine

from finsy.p4schema import P4SchemaCache


def _sigterm_cancel_task(task: asyncio.Task[Any]) -> None:
    task.cancel()


async def _finsy_boilerplate(coro: Coroutine[Any, Any, None]):
    "Wrap main async function and implement boilerplate."
    logging.basicConfig(
        level=logging.INFO,
        format="%(created).03f %(levelname)s %(name)s %(message)s",
    )

    if sys.platform != "win32":
        # Boilerplate to shutdown cleanly upon SIGTERM signal.
        current_task = asyncio.current_task()
        assert current_task is not None
        asyncio.get_running_loop().add_signal_handler(
            signal.SIGTERM,
            _sigterm_cancel_task,
            current_task,
        )

    # Enable caching of P4Info definitions.
    with P4SchemaCache():
        await coro


def run(coro: Coroutine[Any, Any, None]) -> None:
    """`finsy.run` provides a useful wrapper around `asyncio.run`.

    This function implements common boilerplate for running a Finsy application.

    - Set up basic logging to stderr at the INFO log level.
    - Set up a signal handler for SIGTERM that shuts down gracefully.
    - Set up caching of P4Info data so common definitions are re-used.

    Example:

    ```python
    import finsy as fy

    async def main():
        async with fy.Switch("sw", "127.0.0.1:50001") as sw:
            print(sw.p4info)

    if __name__ == "__main__":
        fy.run(main())
    ```

    If you choose to use `asyncio.run` instead, your P4Schema/P4Info objects
    will not be eligible for sharing. You can create your own `P4SchemaCache`
    context manager to implement this.
    """
    try:
        asyncio.run(_finsy_boilerplate(coro))
    except (KeyboardInterrupt, asyncio.CancelledError):
        pass



"Implements the Controller class."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from contextvars import ContextVar
from types import TracebackType
from typing import Any, Iterable, Iterator, final

from typing_extensions import Self

from finsy.futures import CountdownFuture, wait_for_cancel
from finsy.log import LOGGER
from finsy.switch import Switch, SwitchEvent, SwitchFailFastError


@final
class Controller:
    """Represents a collection of P4Runtime switches.

    Each `Switch` in the Controller is identified by its name. Each name must
    be unique.

    ```
    switches = [
        fy.Switch("sw1", "10.0.0.1:50000"),
        fy.Switch("sw2", "10.0.0.2:50000"),
    ]
    controller = fy.Controller(switches)
    await controller.run()
    ```

    A Controller can be running or stopped. There are two ways to run a
    Controller. You can use the `Controller.run()` method, or you can use
    a Controller as a context manager.

    ```
    controller = fy.Controller(switches)
    async with controller:
        # Let controller run for 60 seconds.
        await asyncio.sleep(60)
    ```

    You can `add` or `remove` switches regardless of whether a Controller is
    running or not. If the Controller is not running, adding or removing a Switch is
    instantaneous. If the Controller is running, adding a Switch will start
    it running asynchronously. Removing a Switch will schedule the Switch to stop,
    but defer actual removal until the Switch has stopped asynchronously.

    When a switch starts inside a Controller, it fires the `CONTROLLER_ENTER`
    event. When a switch stops inside a Controller, it fires the `CONTROLLER_LEAVE`
    event.

    A Controller supports these methods to access its contents:

    - len(controller): Return number of switches in the controller.
    - controller[name]: Return the switch with the given name.
    - controller.get(name): Return the switch with the given name, or None if not found.

    You can iterate over the switches in a Controller using a for loop:

    ```
    for switch in controller:
        print(switch.name)
    ```

    Any task or sub-task running inside a controller can retrieve its
    Controller object using the `Controller.current()` method.
    """

    _switches: dict[str, Switch]
    _pending_removal: set[Switch]
    _task_count: CountdownFuture

    _control_task: asyncio.Task[Any] | None = None
    "Keep track of controller's main task."

    def __init__(self, switches: Iterable[Switch] = ()):
        self._switches = {}
        self._pending_removal = set()
        self._task_count = CountdownFuture()

        for switch in switches:
            if switch.name in self._switches:
                raise ValueError(f"Switch named {switch.name!r} already exists")
            self._switches[switch.name] = switch

    @property
    def running(self) -> bool:
        "True if Controller is running."
        return self._control_task is not None

    async def run(self) -> None:
        "Run the controller."
        async with self:
            await wait_for_cancel()

    def stop(self) -> None:
        "Stop the controller if it is running."
        if self._control_task is not None:
            self._control_task.cancel()

    async def __aenter__(self) -> Self:
        "Run the controller as a context manager (see also run())."
        assert not self.running, "Controller.__aenter__ is not re-entrant"
        assert self._task_count.value() == 0
        assert not self._pending_removal

        self._control_task = asyncio.current_task()
        _CONTROLLER.set(self)

        try:
            # Start each switch running.
            for switch in self:
                self._start_switch(switch)
        except Exception:
            self._control_task = None
            _CONTROLLER.set(None)
            raise

        return self

    async def __aexit__(
        self,
        _exc_type: type[BaseException] | None,
        _exc_val: BaseException | None,
        _exc_tb: TracebackType | None,
    ) -> bool | None:
        "Run the controller as a context manager (see also run())."
        assert self.running

        try:
            # Stop all the switches.
            for switch in self:
                self._stop_switch(switch)

            # Wait for switch tasks to finish.
            await self._task_count.wait()

        finally:
            self._control_task = None
            _CONTROLLER.set(None)

    def add(self, switch: Switch) -> None:
        """Add a switch to the controller.

        If the controller is running, tell the switch to start.
        """
        if switch.name in self._switches:
            raise ValueError(f"Switch named {switch.name!r} already exists")

        self._switches[switch.name] = switch
        if self.running:
            self._start_switch(switch)

    def remove(self, switch: Switch) -> asyncio.Event:
        """Remove a switch from the controller.

        If the controller is running, tell the switch to stop and schedule it
        for removal when it fully stops.
        """
        name = switch.name
        if self._switches.get(name, None) is not switch:
            raise ValueError(f"Switch named {name!r} not found")

        del self._switches[name]

        event = asyncio.Event()
        if self.running:
            # When controller is running, event will complete when switch
            # is actually stopped.
            self._stop_switch(switch)
            self._pending_removal.add(switch)

            def _controller_leave(sw: Switch):
                self._pending_removal.discard(sw)
                event.set()

            switch.ee.once(SwitchEvent.CONTROLLER_LEAVE, _controller_leave)  # type: ignore
        else:
            # When controller is not running, event completes immediately.
            event.set()

        return event

    def _start_switch(self, switch: Switch):
        "Start the switch's control task."
        LOGGER.debug("Controller._start_switch: %r", switch)
        assert switch._control_task is None  # pyright: ignore[reportPrivateUsage]

        switch.ee.emit(SwitchEvent.CONTROLLER_ENTER, switch)

        task = asyncio.create_task(switch.run(), name=f"fy:{switch.name}")
        switch._control_task = task  # pyright: ignore[reportPrivateUsage]
        self._task_count.increment()

        def _switch_done(done: asyncio.Task[Any]):
            switch._control_task = None  # pyright: ignore[reportPrivateUsage]
            switch.ee.emit(SwitchEvent.CONTROLLER_LEAVE, switch)
            self._task_count.decrement()

            if not done.cancelled():
                ex = done.exception()
                if ex is not None:
                    if not isinstance(ex, SwitchFailFastError):
                        # The `fail_fast` error has already been logged. If
                        # it's any other error, log it. (There shouldn't be
                        # any other error.)
                        LOGGER.critical(
                            "Controller task %r failed",
                            done.get_name(),
                            exc_info=ex,
                        )
                    # Shutdown the program cleanly due to switch failure.
                    raise SystemExit(99)

        task.add_done_callback(_switch_done)

    def _stop_switch(self, switch: Switch):
        "Stop the switch's control task."
        LOGGER.debug("Controller._stop_switch: %r", switch)

        if switch._control_task is not None:  # pyright: ignore[reportPrivateUsage]
            switch._control_task.cancel()  # pyright: ignore[reportPrivateUsage]

    def __len__(self) -> int:
        "Return switch count."
        return len(self._switches)

    def __iter__(self) -> Iterator[Switch]:
        "Iterate over switches."
        return iter(self._switches.values())

    def __getitem__(self, name: str) -> Switch:
        "Retrieve switch by name."
        return self._switches[name]

    def get(self, name: str) -> Switch | None:
        "Retrieve switch by name, or return None if not found."
        return self._switches.get(name)

    @staticmethod
    def current() -> "Controller":
        "Return the current Controller object."
        result = _CONTROLLER.get()
        if result is None:
            raise RuntimeError("controller does not exist")
        return result


_CONTROLLER: ContextVar[Controller | None] = ContextVar("_CONTROLLER", default=None)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: p4/v1/p4data.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12p4/v1/p4data.proto\x12\x05p4.v1\"\x94\x03\n\x06P4Data\x12\x13\n\tbitstring\x18\x01 \x01(\x0cH\x00\x12!\n\x06varbit\x18\x02 \x01(\x0b\x32\x0f.p4.v1.P4VarbitH\x00\x12\x0e\n\x04\x62ool\x18\x03 \x01(\x08H\x00\x12$\n\x05tuple\x18\x04 \x01(\x0b\x32\x13.p4.v1.P4StructLikeH\x00\x12%\n\x06struct\x18\x05 \x01(\x0b\x32\x13.p4.v1.P4StructLikeH\x00\x12!\n\x06header\x18\x06 \x01(\x0b\x32\x0f.p4.v1.P4HeaderH\x00\x12,\n\x0cheader_union\x18\x07 \x01(\x0b\x32\x14.p4.v1.P4HeaderUnionH\x00\x12,\n\x0cheader_stack\x18\x08 \x01(\x0b\x32\x14.p4.v1.P4HeaderStackH\x00\x12\x37\n\x12header_union_stack\x18\t \x01(\x0b\x32\x19.p4.v1.P4HeaderUnionStackH\x00\x12\x0e\n\x04\x65num\x18\n \x01(\tH\x00\x12\x0f\n\x05\x65rror\x18\x0b \x01(\tH\x00\x12\x14\n\nenum_value\x18\x0c \x01(\x0cH\x00\x42\x06\n\x04\x64\x61ta\"/\n\x08P4Varbit\x12\x11\n\tbitstring\x18\x01 \x01(\x0c\x12\x10\n\x08\x62itwidth\x18\x02 \x01(\x05\".\n\x0cP4StructLike\x12\x1e\n\x07members\x18\x01 \x03(\x0b\x32\r.p4.v1.P4Data\"0\n\x08P4Header\x12\x10\n\x08is_valid\x18\x01 \x01(\x08\x12\x12\n\nbitstrings\x18\x02 \x03(\x0c\"Q\n\rP4HeaderUnion\x12\x19\n\x11valid_header_name\x18\x01 \x01(\t\x12%\n\x0cvalid_header\x18\x02 \x01(\x0b\x32\x0f.p4.v1.P4Header\"1\n\rP4HeaderStack\x12 \n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x0f.p4.v1.P4Header\";\n\x12P4HeaderUnionStack\x12%\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x14.p4.v1.P4HeaderUnionB&Z$github.com/p4lang/p4runtime/go/p4/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'p4.v1.p4data_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z$github.com/p4lang/p4runtime/go/p4/v1'
  _globals['_P4DATA']._serialized_start=30
  _globals['_P4DATA']._serialized_end=434
  _globals['_P4VARBIT']._serialized_start=436
  _globals['_P4VARBIT']._serialized_end=483
  _globals['_P4STRUCTLIKE']._serialized_start=485
  _globals['_P4STRUCTLIKE']._serialized_end=531
  _globals['_P4HEADER']._serialized_start=533
  _globals['_P4HEADER']._serialized_end=581
  _globals['_P4HEADERUNION']._serialized_start=583
  _globals['_P4HEADERUNION']._serialized_end=664
  _globals['_P4HEADERSTACK']._serialized_start=666
  _globals['_P4HEADERSTACK']._serialized_end=715
  _globals['_P4HEADERUNIONSTACK']._serialized_start=717
  _globals['_P4HEADERUNIONSTACK']._serialized_end=776
# @@protoc_insertion_point(module_scope)
